"""Compound tensor algebra operations."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.algebra import Conj, Operator
from ufl.constantvalue import Zero
from ufl.core.expr import ufl_err_str
from ufl.core.ufl_type import ufl_type
from ufl.index_combination_utils import merge_nonoverlapping_indices
from ufl.precedence import parstr
from ufl.sorting import sorted_expr

# Algebraic operations on tensors:
# FloatValues:
#   dot(a,b)      = a*b
#   inner(a,b)    = a*b
#   outer(a,b)    = a*b
# Vectors:
#   dot(u,v)      = u_i v_i
#   inner(u,v)    = u_i v_i
#   outer(u,v)    = A | A_ij = u_i v_j
# Matrices:
#   dot(A,B)      = C | C_ij = A_{ik} B_{kj}
#   inner(A,B)    = A_{ij} B_{ij}
#   outer(A,B)    = C | C_ijkl = A_ij B_kl
# Combined:
#   dot(A,u)      = v | v_i = A_{ik} u_k
#   inner(A,u)    = v | v_i = A_{ik} u_k
#   outer(A,u)    = C | C_ijk = B_ij u_k
#   dot(u,B)      = v | v_i = u_k B_{ki}
#   inner(u,B)    = v | v_i = u_k B_{ki}
#   outer(u,B)    = C | C_ijk = u_i B_jk
#
# Argument requirements:
#   dot(x,y):   last index of x has same dimension as first index of y
#   inner(x,y): shape of x equals the shape of y


# --- Classes representing compound tensor algebra operations ---


@ufl_type(is_abstract=True)
class CompoundTensorOperator(Operator):
    """Compount tensor operator."""

    __slots__ = ()

    def __init__(self, operands):
        """Initialise."""
        Operator.__init__(self, operands)


# TODO: Use this and make Sum handle scalars only?
#       This would simplify some algorithms. The only
#       problem is we can't use + in many algorithms because
#       this type should be expanded by expand_compounds.
# class TensorSum(CompoundTensorOperator):
#     "Sum of nonscalar expressions."
#     pass

# TODO: Use this similarly to TensorSum?
#       This would simplify some algorithms. The only
#       problem is we can't use / in many algorithms because
#       this type should be expanded by expand_compounds.
# class TensorDivision(CompoundTensorOperator):
#     "Division of nonscalar expression with a scalar expression."
#     pass

# TODO: Use this similarly to TensorSum?
#       This would simplify some algorithms. The only
#       problem is we can't use * in many algorithms because
#       this type should be expanded by expand_compounds.
# class MatrixProduct(CompoundTensorOperator):
#     "Product of a matrix with a matrix or vector."
#     pass

# TODO: Use this similarly to TensorSum?
#       This would simplify some algorithms. The only
#       problem is we can't use abs in many algorithms because
#       this type should be expanded by expand_compounds.
# class TensorAbs(CompoundTensorOperator):
#     "Absolute value of nonscalar expression."
#     pass


@ufl_type(is_shaping=True, num_ops=1, inherit_indices_from_operand=0)
class Transposed(CompoundTensorOperator):
    """Transposed tensor."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Transposed."""
        if isinstance(A, Zero):
            a, b = A.ufl_shape
            return Zero((b, a), A.ufl_free_indices, A.ufl_index_dimensions)
        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))
        if len(A.ufl_shape) != 2:
            raise ValueError("Transposed is only defined for rank 2 tensors.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        s = self.ufl_operands[0].ufl_shape
        return (s[1], s[0])

    def __str__(self):
        """Format as a string."""
        return "%s^T" % parstr(self.ufl_operands[0], self)


@ufl_type(num_ops=2)
class Outer(CompoundTensorOperator):
    """Outer."""

    __slots__ = ("ufl_free_indices", "ufl_index_dimensions")

    def __new__(cls, a, b):
        """Create new Outer."""
        ash, bsh = a.ufl_shape, b.ufl_shape
        if isinstance(a, Zero) or isinstance(b, Zero):
            fi, fid = merge_nonoverlapping_indices(a, b)
            return Zero(ash + bsh, fi, fid)
        if ash == () or bsh == ():
            return Conj(a) * b
        return CompoundTensorOperator.__new__(cls)

    def __init__(self, a, b):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (a, b))
        fi, fid = merge_nonoverlapping_indices(a, b)
        self.ufl_free_indices = fi
        self.ufl_index_dimensions = fid

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape + self.ufl_operands[1].ufl_shape

    def __str__(self):
        """Format as a string."""
        return "%s (X) %s" % (
            parstr(self.ufl_operands[0], self),
            parstr(self.ufl_operands[1], self),
        )


@ufl_type(num_ops=2)
class Inner(CompoundTensorOperator):
    """Inner."""

    __slots__ = ("ufl_free_indices", "ufl_index_dimensions")

    def __new__(cls, a, b):
        """Create new Inner."""
        # Checks
        ash, bsh = a.ufl_shape, b.ufl_shape
        if ash != bsh:
            raise ValueError(f"Shapes do not match: {ufl_err_str(a)} and {ufl_err_str(b)}")

        # Simplification
        if isinstance(a, Zero) or isinstance(b, Zero):
            fi, fid = merge_nonoverlapping_indices(a, b)
            return Zero((), fi, fid)
        elif ash == ():
            return a * Conj(b)
        # sort operands for unique representation,
        # must be independent of various counts etc.
        # as explained in cmp_expr
        if (a, b) != tuple(sorted_expr((a, b))):
            return Conj(Inner(b, a))

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, a, b):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (a, b))

        fi, fid = merge_nonoverlapping_indices(a, b)
        self.ufl_free_indices = fi
        self.ufl_index_dimensions = fid

    ufl_shape = ()

    def __str__(self):
        """Format as a string."""
        return "%s : %s" % (parstr(self.ufl_operands[0], self), parstr(self.ufl_operands[1], self))


@ufl_type(num_ops=2)
class Dot(CompoundTensorOperator):
    """Dot."""

    __slots__ = ("ufl_free_indices", "ufl_index_dimensions")

    def __new__(cls, a, b):
        """Create new Dot."""
        ash = a.ufl_shape
        bsh = b.ufl_shape
        ar, br = len(ash), len(bsh)
        scalar = ar == 0 and br == 0

        # Checks
        if not ((ar >= 1 and br >= 1) or scalar):
            raise ValueError(
                "Dot product requires non-scalar arguments, "
                f"got arguments with ranks {ar} and {br}."
            )
        if not (scalar or ash[-1] == bsh[0]):
            raise ValueError("Dimension mismatch in dot product.")

        # Simplification
        if isinstance(a, Zero) or isinstance(b, Zero):
            shape = ash[:-1] + bsh[1:]
            fi, fid = merge_nonoverlapping_indices(a, b)
            return Zero(shape, fi, fid)
        elif scalar:  # TODO: Move this to def dot()?
            return a * b

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, a, b):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (a, b))
        fi, fid = merge_nonoverlapping_indices(a, b)
        self.ufl_free_indices = fi
        self.ufl_index_dimensions = fid

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape[:-1] + self.ufl_operands[1].ufl_shape[1:]

    def __str__(self):
        """Format as a string."""
        return "%s . %s" % (parstr(self.ufl_operands[0], self), parstr(self.ufl_operands[1], self))


@ufl_type(is_index_free=True, num_ops=1)
class Perp(CompoundTensorOperator):
    """Perp."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Perp."""
        sh = A.ufl_shape

        # Checks
        if not len(sh) == 1:
            raise ValueError(f"Perp requires arguments of rank 1, got {ufl_err_str(A)}")
        if not sh[0] == 2:
            raise ValueError(f"Perp can only work on 2D vectors, got {ufl_err_str(A)}")

        # Simplification
        if isinstance(A, Zero):
            return Zero(sh, A.ufl_free_indices, A.ufl_index_dimensions)

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

    ufl_shape = (2,)

    def __str__(self):
        """Format as a string."""
        return "perp(%s)" % self.ufl_operands[0]


@ufl_type(num_ops=2)
class Cross(CompoundTensorOperator):
    """Cross."""

    __slots__ = ("ufl_free_indices", "ufl_index_dimensions")

    def __new__(cls, a, b):
        """Create new Cross."""
        ash = a.ufl_shape
        bsh = b.ufl_shape

        # Checks
        if not (len(ash) == 1 and ash == bsh):
            raise ValueError(
                f"Cross product requires arguments of rank 1, got {ufl_err_str(a)} "
                f"and {ufl_err_str(b)}."
            )

        # Simplification
        if isinstance(a, Zero) or isinstance(b, Zero):
            fi, fid = merge_nonoverlapping_indices(a, b)
            return Zero(ash, fi, fid)

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, a, b):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (a, b))
        fi, fid = merge_nonoverlapping_indices(a, b)
        self.ufl_free_indices = fi
        self.ufl_index_dimensions = fid

    ufl_shape = (3,)

    def __str__(self):
        """Format as a string."""
        return "%s x %s" % (parstr(self.ufl_operands[0], self), parstr(self.ufl_operands[1], self))


@ufl_type(num_ops=1, inherit_indices_from_operand=0)
class Trace(CompoundTensorOperator):
    """Trace."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Trace."""
        # Checks
        if len(A.ufl_shape) != 2:
            raise ValueError("Trace of tensor with rank != 2 is undefined.")

        # Simplification
        if isinstance(A, Zero):
            return Zero((), A.ufl_free_indices, A.ufl_index_dimensions)

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

    ufl_shape = ()

    def __str__(self):
        """Format as a string."""
        return "tr(%s)" % self.ufl_operands[0]


@ufl_type(is_scalar=True, num_ops=1)
class Determinant(CompoundTensorOperator):
    """Determinant."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Determinant."""
        sh = A.ufl_shape
        r = len(sh)
        Afi = A.ufl_free_indices

        # Checks
        if r not in (0, 2):
            raise ValueError("Determinant of tensor with rank != 2 is undefined.")
        if r == 2 and sh[0] != sh[1]:
            raise ValueError("Cannot take determinant of rectangular rank 2 tensor.")
        if Afi:
            raise ValueError("Not expecting free indices in determinant.")

        # Simplification
        if isinstance(A, Zero):
            return Zero((), Afi, A.ufl_index_dimensions)
        if r == 0:
            return A

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

    def __str__(self):
        """Format as a string."""
        return "det(%s)" % self.ufl_operands[0]


# TODO: Drop Inverse and represent it as product of Determinant and
# Cofactor?
@ufl_type(is_index_free=True, num_ops=1)
class Inverse(CompoundTensorOperator):
    """Inverse."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Inverse."""
        sh = A.ufl_shape
        r = len(sh)

        # Checks
        if A.ufl_free_indices:
            raise ValueError("Not expecting free indices in Inverse.")
        if isinstance(A, Zero):
            raise ValueError("Division by zero!")

        # Simplification
        if r == 0:
            return 1 / A

        # More checks
        if r != 2:
            raise ValueError("Inverse of tensor with rank != 2 is undefined.")
        if sh[0] != sh[1]:
            raise ValueError(f"Cannot take inverse of rectangular matrix with dimensions {sh}.")

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape

    def __str__(self):
        """Format as a string."""
        return "%s^-1" % parstr(self.ufl_operands[0], self)


@ufl_type(is_index_free=True, num_ops=1)
class Cofactor(CompoundTensorOperator):
    """Cofactor."""

    __slots__ = ()

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

        # Checks
        sh = A.ufl_shape
        if len(sh) != 2:
            raise ValueError("Cofactor of tensor with rank != 2 is undefined.")
        if sh[0] != sh[1]:
            raise ValueError(f"Cannot take cofactor of rectangular matrix with dimensions {sh}.")
        if A.ufl_free_indices:
            raise ValueError("Not expecting free indices in Cofactor.")
        if isinstance(A, Zero):
            raise ValueError("Cannot take cofactor of zero matrix.")

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape

    def __str__(self):
        """Format as a string."""
        return "cofactor(%s)" % self.ufl_operands[0]


@ufl_type(num_ops=1, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class Deviatoric(CompoundTensorOperator):
    """Deviatoric."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Deviatoric."""
        sh = A.ufl_shape

        # Checks
        if len(sh) != 2:
            raise ValueError("Deviatoric part of tensor with rank != 2 is undefined.")
        if sh[0] != sh[1]:
            raise ValueError(
                f"Cannot take deviatoric part of rectangular matrix with dimensions {sh}."
            )
        if A.ufl_free_indices:
            raise ValueError("Not expecting free indices in Deviatoric.")

        # Simplification
        if isinstance(A, Zero):
            return Zero(A.ufl_shape, A.ufl_free_indices, A.ufl_index_dimensions)

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

    def __str__(self):
        """Format as a string."""
        return "dev(%s)" % self.ufl_operands[0]


@ufl_type(num_ops=1, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class Skew(CompoundTensorOperator):
    """Skew."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Skew."""
        sh = A.ufl_shape
        Afi = A.ufl_free_indices

        # Checks
        if len(sh) != 2:
            raise ValueError("Skew symmetric part of tensor with rank != 2 is undefined.")
        if sh[0] != sh[1]:
            raise ValueError(f"Cannot take skew part of rectangular matrix with dimensions {sh}.")
        if Afi:
            raise ValueError("Not expecting free indices in Skew.")

        # Simplification
        if isinstance(A, Zero):
            return Zero(A.ufl_shape, Afi, A.ufl_index_dimensions)

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

    def __str__(self):
        """Format as a string."""
        return "skew(%s)" % self.ufl_operands[0]


@ufl_type(num_ops=1, inherit_shape_from_operand=0, inherit_indices_from_operand=0)
class Sym(CompoundTensorOperator):
    """Sym."""

    __slots__ = ()

    def __new__(cls, A):
        """Create new Sym."""
        sh = A.ufl_shape
        Afi = A.ufl_free_indices

        # Checks
        if len(sh) != 2:
            raise ValueError("Symmetric part of tensor with rank != 2 is undefined.")
        if sh[0] != sh[1]:
            raise ValueError(
                f"Cannot take symmetric part of rectangular matrix with dimensions {sh}."
            )
        if Afi:
            raise ValueError("Not expecting free indices in Sym.")

        # Simplification
        if isinstance(A, Zero):
            return Zero(A.ufl_shape, Afi, A.ufl_index_dimensions)

        return CompoundTensorOperator.__new__(cls)

    def __init__(self, A):
        """Initialise."""
        CompoundTensorOperator.__init__(self, (A,))

    def __str__(self):
        """Format as a string."""
        return f"sym({self.ufl_operands[0]})"



import cmath

import pytest

from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    as_tensor,
    as_ufl,
    atan,
    conditional,
    conj,
    cos,
    cosh,
    dot,
    dx,
    exp,
    ge,
    grad,
    gt,
    imag,
    inner,
    le,
    ln,
    lt,
    max_value,
    min_value,
    outer,
    real,
    sin,
    sqrt,
    triangle,
)
from ufl.algebra import Conj, Imag, Real
from ufl.algorithms import estimate_total_polynomial_degree
from ufl.algorithms.apply_algebra_lowering import apply_algebra_lowering
from ufl.algorithms.comparison_checker import ComplexComparisonError, do_comparison_check
from ufl.algorithms.formtransformations import compute_form_adjoint
from ufl.algorithms.remove_complex_nodes import remove_complex_nodes
from ufl.constantvalue import ComplexValue, Zero
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_conj(self):
    z1 = ComplexValue(1 + 2j)
    z2 = ComplexValue(1 - 2j)

    assert z1 == Conj(z2)
    assert z2 == Conj(z1)


def test_real(self):
    z0 = Zero()
    z1 = as_ufl(1.0)
    z2 = ComplexValue(1j)
    z3 = ComplexValue(1 + 1j)
    assert Real(z1) == z1
    assert Real(z3) == z1
    assert Real(z2) == z0


def test_imag(self):
    z0 = Zero()
    z1 = as_ufl(1.0)
    z2 = as_ufl(1j)
    z3 = ComplexValue(1 + 1j)

    assert Imag(z2) == z1
    assert Imag(z3) == z1
    assert Imag(z1) == z0


def test_compute_form_adjoint(self):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    u = TrialFunction(space)
    v = TestFunction(space)

    a = inner(grad(u), grad(v)) * dx

    assert compute_form_adjoint(a) == conj(inner(grad(v), grad(u))) * dx


def test_complex_algebra(self):
    z1 = ComplexValue(1j)
    z2 = ComplexValue(1 + 1j)

    # Remember that ufl.algebra functions return ComplexValues, but
    # ufl.mathfunctions return complex Python scalar
    # Any operations with a ComplexValue and a complex Python scalar
    # promote to ComplexValue
    assert z1 * z2 == ComplexValue(-1 + 1j)
    assert z2 / z1 == ComplexValue(1 - 1j)
    assert pow(z2, z1) == ComplexValue((1 + 1j) ** 1j)
    assert sqrt(z2) * as_ufl(1) == ComplexValue(cmath.sqrt(1 + 1j))
    assert (sin(z2) + cosh(z2) - atan(z2)) * z1 == ComplexValue(
        (cmath.sin(1 + 1j) + cmath.cosh(1 + 1j) - cmath.atan(1 + 1j)) * 1j
    )
    assert (abs(z2) - ln(z2)) / exp(z1) == ComplexValue(
        (abs(1 + 1j) - cmath.log(1 + 1j)) / cmath.exp(1j)
    )


def test_automatic_simplification(self):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    assert inner(u, v) == u * conj(v)
    assert dot(u, v) == u * v
    assert outer(u, v) == conj(u) * v


def test_apply_algebra_lowering_complex(self):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)

    gv = grad(v)
    gu = grad(u)

    a = dot(gu, gv)
    b = inner(gv, gu)
    c = outer(gu, gv)

    lowered_a = apply_algebra_lowering(a)
    lowered_b = apply_algebra_lowering(b)
    lowered_c = apply_algebra_lowering(c)
    lowered_a_index = lowered_a.index()
    lowered_b_index = lowered_b.index()
    lowered_c_indices = lowered_c.indices()

    assert lowered_a == gu[lowered_a_index] * gv[lowered_a_index]
    assert lowered_b == gv[lowered_b_index] * conj(gu[lowered_b_index])
    assert lowered_c == as_tensor(
        conj(gu[lowered_c_indices[0]]) * gv[lowered_c_indices[1]],
        (lowered_c_indices[0],) + (lowered_c_indices[1],),
    )


def test_remove_complex_nodes(self):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    u = TrialFunction(space)
    v = TestFunction(space)
    f = Coefficient(space)

    a = conj(v)
    b = real(u)
    c = imag(f)
    d = conj(real(v)) * imag(conj(u))

    assert remove_complex_nodes(a) == v
    assert remove_complex_nodes(b) == u
    with pytest.raises(BaseException):
        remove_complex_nodes(c)
    with pytest.raises(BaseException):
        remove_complex_nodes(d)


def test_comparison_checker(self):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    u = TrialFunction(space)
    v = TestFunction(space)

    a = conditional(ge(abs(u), imag(v)), u, v)
    b = conditional(le(sqrt(abs(u)), imag(v)), as_ufl(1), as_ufl(1j))
    c = conditional(gt(abs(u), pow(imag(v), 0.5)), sin(u), cos(v))
    d = conditional(lt(as_ufl(-1), as_ufl(1)), u, v)
    e = max_value(as_ufl(0), real(u))
    f = min_value(sin(u), cos(v))
    g = min_value(sin(pow(u, 3)), cos(abs(v)))

    assert do_comparison_check(a) == conditional(ge(real(abs(u)), real(imag(v))), u, v)
    with pytest.raises(ComplexComparisonError):
        b = do_comparison_check(b)
    with pytest.raises(ComplexComparisonError):
        c = do_comparison_check(c)
    assert do_comparison_check(d) == conditional(lt(real(as_ufl(-1)), real(as_ufl(1))), u, v)
    assert do_comparison_check(e) == max_value(real(as_ufl(0)), real(real(u)))
    assert do_comparison_check(f) == min_value(real(sin(u)), real(cos(v)))
    assert do_comparison_check(g) == min_value(real(sin(pow(u, 3))), real(cos(abs(v))))


def test_complex_degree_handling(self):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 3, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)

    a = conj(v)
    b = imag(v)
    c = real(v)

    assert estimate_total_polynomial_degree(a) == 3
    assert estimate_total_polynomial_degree(b) == 3
    assert estimate_total_polynomial_degree(c) == 3



__authors__ = "Martin Sandve Alnæs"
__date__ = "2011-04-14"

import numpy as np

from ufl import PermutationSymbol, as_matrix, as_vector, indices, product
from ufl.classes import Indexed
from ufl.constantvalue import ComplexValue, FloatValue, IntValue, Zero, as_ufl


def test_zero(self):
    z1 = Zero(())

    assert z1 == z1
    assert int(z1) == 0
    assert float(z1) == 0.0
    assert complex(z1) == 0.0 + 0.0j
    self.assertNotEqual(z1, 1.0)
    self.assertFalse(z1)

    # If zero() == 0 is to be allowed, it must not have the same hash or
    # it will collide with 0 as key in dicts...
    self.assertNotEqual(hash(z1), hash(0.0))
    self.assertNotEqual(hash(z1), hash(0))


def test_float(self):
    f1 = as_ufl(1)
    f2 = as_ufl(1.0)
    f3 = FloatValue(1)
    f4 = FloatValue(1.0)
    f5 = 3 - FloatValue(1) - 1
    f6 = 3 * FloatValue(2) / 6
    f7 = as_ufl(np.ones((1,), dtype="d")[0])

    assert f1 == f1
    self.assertNotEqual(f1, f2)  # IntValue vs FloatValue, == compares representations!
    assert f2 == f3
    assert f2 == f4
    assert f2 == f5
    assert f2 == f6
    assert f2 == f7


def test_int(self):
    f1 = as_ufl(1)
    f2 = as_ufl(1.0)
    f3 = IntValue(1)
    f4 = IntValue(1.0)
    f5 = 3 - IntValue(1) - 1
    f6 = 3 * IntValue(2) / 6
    f7 = as_ufl(np.ones((1,), dtype="int")[0])

    assert f1 == f1
    self.assertNotEqual(f1, f2)  # IntValue vs FloatValue, == compares representations!
    assert f1 == f3
    assert f1 == f4
    assert f1 == f5
    assert f2 == f6  # Division produces a FloatValue
    assert f1 == f7


def test_complex(self):
    f1 = as_ufl(1 + 1j)
    f2 = as_ufl(1)
    f3 = as_ufl(1j)
    f4 = ComplexValue(1 + 1j)
    f5 = ComplexValue(1.0 + 1.0j)
    f6 = as_ufl(1.0)
    f7 = as_ufl(1.0j)
    f8 = as_ufl(np.array([1 + 1j], dtype="complex")[0])

    assert f1 == f1
    assert f1 == f4
    assert f1 == f5  # ComplexValue uses floats
    assert f1 == f2 + f3  # Type promotion of IntValue to ComplexValue with arithmetic
    assert f4 == f2 + f3
    assert f5 == f2 + f3
    assert f4 == f5
    assert f6 + f7 == f2 + f3
    assert f4 == f8


def test_scalar_sums(self):
    n = 10
    s = [as_ufl(i) for i in range(n)]

    for i in range(n):
        self.assertNotEqual(s[i], i + 1)

    for i in range(n):
        assert s[i] == i

    for i in range(n):
        assert 0 + s[i] == i

    for i in range(n):
        assert s[i] + 0 == i

    for i in range(n):
        assert 0 + s[i] + 0 == i

    for i in range(n):
        assert 1 + s[i] - 1 == i

    assert s[1] + s[1] == 2
    assert s[1] + s[2] == 3
    assert s[1] + s[2] + s[3] == s[6]
    assert s[5] - s[2] == 3
    assert 1 * s[5] == 5
    assert 2 * s[5] == 10
    assert s[6] / 3 == 2


def test_identity(self):
    pass  # FIXME


def test_permutation_symbol_3(self):
    e = PermutationSymbol(3)
    assert e.ufl_shape == (3, 3, 3)
    assert eval(repr(e)) == e
    for i in range(3):
        for j in range(3):
            for k in range(3):
                value = (j - i) * (k - i) * (k - j) / 2
                self.assertEqual(e[i, j, k], value)
    i, j, k = indices(3)
    self.assertIsInstance(e[i, j, k], Indexed)
    x = (0, 0, 0)
    self.assertEqual((e[i, j, k] * e[i, j, k])(x), 6)


def test_permutation_symbol_n(self):
    for n in range(2, 5):  # tested with upper limit 7, but evaluation is a bit slow then
        e = PermutationSymbol(n)
        assert e.ufl_shape == (n,) * n
        assert eval(repr(e)) == e

        ii = indices(n)
        x = (0,) * n
        nfac = product(m for m in range(1, n + 1))
        assert (e[ii] * e[ii])(x) == nfac


def test_unit_dyads(self):
    from ufl.tensors import unit_matrices, unit_vectors

    ei, ej = unit_vectors(2)
    self.assertEqual(as_vector((1, 0)), ei)
    self.assertEqual(as_vector((0, 1)), ej)
    eii, eij, eji, ejj = unit_matrices(2)
    self.assertEqual(as_matrix(((1, 0), (0, 0))), eii)
    self.assertEqual(as_matrix(((0, 1), (0, 0))), eij)
    self.assertEqual(as_matrix(((0, 0), (1, 0))), eji)
    self.assertEqual(as_matrix(((0, 0), (0, 1))), ejj)



"""Averaging operations."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from ufl.constantvalue import ConstantValue
from ufl.core.operator import Operator
from ufl.core.ufl_type import ufl_type


@ufl_type(
    inherit_shape_from_operand=0, inherit_indices_from_operand=0, num_ops=1, is_evaluation=True
)
class CellAvg(Operator):
    """Cell average."""

    __slots__ = ()

    def __new__(cls, f):
        """Create a new CellAvg."""
        if isinstance(f, ConstantValue):
            return f
        return super(CellAvg, cls).__new__(cls)

    def __init__(self, f):
        """Initialise."""
        Operator.__init__(self, (f,))

    @property
    def ufl_shape(self):
        """Get the UFL shape."""
        return self.ufl_operands[0].ufl_shape

    def evaluate(self, x, mapping, component, index_values):
        """Performs an approximate symbolic evaluation, since we don't have a cell."""
        return self.ufl_operands[0].evaluate(x, mapping, component, index_values)

    def __str__(self):
        """Format as a string."""
        return f"cell_avg({self.ufl_operands[0]})"


@ufl_type(
    inherit_shape_from_operand=0, inherit_indices_from_operand=0, num_ops=1, is_evaluation=True
)
class FacetAvg(Operator):
    """Facet average."""

    __slots__ = ()

    def __new__(cls, f):
        """Create a new FacetAvg."""
        if isinstance(f, ConstantValue):
            return f
        return super(FacetAvg, cls).__new__(cls)

    def __init__(self, f):
        """Initialise."""
        Operator.__init__(self, (f,))

    @property
    def ufl_shape(self):
        """Return the UFL shape."""
        return self.ufl_operands[0].ufl_shape

    def evaluate(self, x, mapping, component, index_values):
        """Performs an approximate symbolic evaluation, since we dont have a cell."""
        return self.ufl_operands[0].evaluate(x, mapping, component, index_values)

    def __str__(self):
        """Format as a string."""
        return f"facet_avg({self.ufl_operands[0]})"



__authors__ = "Martin Sandve Alnæs"
__date__ = "2009-02-17 -- 2009-02-17"

from itertools import chain

from ufl import (
    CellDiameter,
    CellVolume,
    Circumradius,
    Coefficient,
    Constant,
    FacetArea,
    FacetNormal,
    FunctionSpace,
    Identity,
    Index,
    Jacobian,
    JacobianInverse,
    Mesh,
    SpatialCoordinate,
    TestFunction,
    TrialFunction,
    acos,
    as_matrix,
    as_tensor,
    as_vector,
    asin,
    atan,
    conditional,
    cos,
    derivative,
    diff,
    dot,
    dx,
    exp,
    i,
    indices,
    inner,
    interval,
    j,
    k,
    ln,
    lt,
    nabla_grad,
    outer,
    quadrilateral,
    replace,
    sign,
    sin,
    split,
    sqrt,
    tan,
    tetrahedron,
    triangle,
    variable,
    zero,
)
from ufl.algorithms import compute_form_data, expand_indices, strip_variables
from ufl.algorithms.apply_algebra_lowering import apply_algebra_lowering
from ufl.algorithms.apply_derivatives import apply_derivatives
from ufl.algorithms.apply_geometry_lowering import apply_geometry_lowering
from ufl.classes import Indexed, MultiIndex, ReferenceGrad
from ufl.constantvalue import as_ufl
from ufl.domain import extract_unique_domain
from ufl.finiteelement import FiniteElement, MixedElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1, L2


def assertEqualBySampling(actual, expected):
    ad = compute_form_data(actual * dx)
    a = ad.preprocessed_form.integrals_by_type("cell")[0].integrand()
    bd = compute_form_data(expected * dx)
    b = bd.preprocessed_form.integrals_by_type("cell")[0].integrand()

    assert [ad.function_replace_map[ac] for ac in ad.reduced_coefficients] == [
        bd.function_replace_map[bc] for bc in bd.reduced_coefficients
    ]

    n = ad.num_coefficients

    def make_value(c):
        if isinstance(c, Coefficient):
            z = 0.3
            m = c.count()
        else:
            z = 0.7
            m = c.number()
        if c.ufl_shape == ():
            return z * (0.1 + 0.9 * m / n)
        elif len(c.ufl_shape) == 1:
            return tuple((z * (j + 0.1 + 0.9 * m / n) for j in range(c.ufl_shape[0])))
        else:
            raise NotImplementedError("Tensor valued expressions not supported here.")

    amapping = dict(
        (c, make_value(c))
        for c in chain(ad.original_form.coefficients(), ad.original_form.arguments())
    )
    bmapping = dict(
        (c, make_value(c))
        for c in chain(bd.original_form.coefficients(), bd.original_form.arguments())
    )
    adomain = extract_unique_domain(actual)
    bdomain = extract_unique_domain(expected)
    acell = adomain.ufl_cell()
    bcell = bdomain.ufl_cell()
    assert acell == bcell
    if adomain.geometric_dimension() == 1:
        x = (0.3,)
    elif adomain.geometric_dimension() == 2:
        x = (0.3, 0.4)
    elif adomain.geometric_dimension() == 3:
        x = (0.3, 0.4, 0.5)
    av = a(x, amapping)
    bv = b(x, bmapping)

    if not av == bv:
        print("Tried to sample expressions to compare but failed:")
        print()
        print((str(a)))
        print(av)
        print()
        print((str(b)))
        print(bv)
        print()

    assert av == bv


def _test(self, f, df):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)

    v = TestFunction(space)
    u = TrialFunction(space)
    w = Coefficient(space)
    xv = (0.3, 0.7)
    uv = 7.0
    vv = 13.0
    wv = 11.0

    x = xv
    mapping = {v: vv, u: uv, w: wv}

    dfv1 = derivative(f(w), w, v)
    dfv2 = df(w, v)
    dfv1 = dfv1(x, mapping)
    dfv2 = dfv2(x, mapping)
    assert dfv1 == dfv2

    dfv1 = derivative(f(7 * w), w, v)
    dfv2 = 7 * df(7 * w, v)
    dfv1 = dfv1(x, mapping)
    dfv2 = dfv2(x, mapping)
    assert dfv1 == dfv2


# --- Literals


def testScalarLiteral(self):
    def f(w):
        return as_ufl(1)

    def df(w, v):
        return zero()

    _test(self, f, df)


def testIdentityLiteral(self):
    def f(w):
        return Identity(2)[i, i]

    def df(w, v):
        return zero()

    _test(self, f, df)


# --- Form arguments


def testCoefficient(self):
    def f(w):
        return w

    def df(w, v):
        return v

    _test(self, f, df)


def testArgument(self):
    def f(w):
        return TestFunction(
            FunctionSpace(
                Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)),
                FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1),
            )
        )

    def df(w, v):
        return zero()

    _test(self, f, df)


# --- Geometry


def testSpatialCoordinate(self):
    def f(w):
        return SpatialCoordinate(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
        )[0]

    def df(w, v):
        return zero()

    _test(self, f, df)


def testFacetNormal(self):
    def f(w):
        return FacetNormal(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
        )[0]

    def df(w, v):
        return zero()

    _test(self, f, df)


def testFacetArea(self):
    def f(w):
        return FacetArea(Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)))

    def df(w, v):
        return zero()

    _test(self, f, df)


def testCellDiameter(self):
    def f(w):
        return CellDiameter(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
        )

    def df(w, v):
        return zero()

    _test(self, f, df)


def testCircumradius(self):
    def f(w):
        return Circumradius(
            Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
        )

    def df(w, v):
        return zero()

    _test(self, f, df)


def testCellVolume(self):
    def f(w):
        return CellVolume(Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)))

    def df(w, v):
        return zero()

    _test(self, f, df)


# --- Basic operators


def testSum(self):
    def f(w):
        return w + 1

    def df(w, v):
        return v

    _test(self, f, df)


def testProduct(self):
    def f(w):
        return 3 * w

    def df(w, v):
        return 3 * v

    _test(self, f, df)


def testPower(self):
    def f(w):
        return w**3

    def df(w, v):
        return 3 * w**2 * v

    _test(self, f, df)


def testDivision(self):
    def f(w):
        return w / 3.0

    def df(w, v):
        return v / 3.0

    _test(self, f, df)


def testDivision2(self):
    def f(w):
        return 3.0 / w

    def df(w, v):
        return -3.0 * v / w**2

    _test(self, f, df)


def testExp(self):
    def f(w):
        return exp(w)

    def df(w, v):
        return v * exp(w)

    _test(self, f, df)


def testLn(self):
    def f(w):
        return ln(w)

    def df(w, v):
        return v / w

    _test(self, f, df)


def testCos(self):
    def f(w):
        return cos(w)

    def df(w, v):
        return -v * sin(w)

    _test(self, f, df)


def testSin(self):
    def f(w):
        return sin(w)

    def df(w, v):
        return v * cos(w)

    _test(self, f, df)


def testTan(self):
    def f(w):
        return tan(w)

    def df(w, v):
        return v * 2.0 / (cos(2.0 * w) + 1.0)

    _test(self, f, df)


def testAcos(self):
    def f(w):
        return acos(w / 1000)

    def df(w, v):
        return -(v / 1000) / sqrt(1.0 - (w / 1000) ** 2)

    _test(self, f, df)


def testAsin(self):
    def f(w):
        return asin(w / 1000)

    def df(w, v):
        return (v / 1000) / sqrt(1.0 - (w / 1000) ** 2)

    _test(self, f, df)


def testAtan(self):
    def f(w):
        return atan(w)

    def df(w, v):
        return v / (1.0 + w**2)

    _test(self, f, df)


# FIXME: Add the new erf and bessel_*

# --- Abs and conditionals


def testAbs(self):
    def f(w):
        return abs(w)

    def df(w, v):
        return sign(w) * v

    _test(self, f, df)


def testConditional(self):  # This will fail without bugfix in derivative
    def cond(w):
        return lt(w, 1.0)

    def f(w):
        return conditional(cond(w), 2 * w, 3 * w)

    def df(w, v):
        return conditional(cond(w), 1, 0) * 2 * v + conditional(cond(w), 0, 1) * 3 * v

    _test(self, f, df)


# --- Tensor algebra basics


def testIndexSum(self):
    def f(w):
        # 3*w + 4*w**2 + 5*w**3
        a = as_vector((w, w**2, w**3))
        b = as_vector((3, 4, 5))
        (i,) = indices(1)
        return a[i] * b[i]

    def df(w, v):
        return 3 * v + 4 * 2 * w * v + 5 * 3 * w**2 * v

    _test(self, f, df)


def testListTensor(self):
    v = variable(as_ufl(42))
    f = as_tensor(
        (
            ((0, 0), (0, 0)),
            ((v, 2 * v), (0, 0)),
            ((v**2, 1), (2, v / 2)),
        )
    )
    assert f.ufl_shape == (3, 2, 2)
    g = as_tensor(
        (
            ((0, 0), (0, 0)),
            ((1, 2), (0, 0)),
            ((84, 0), (0, 0.5)),
        )
    )
    assert g.ufl_shape == (3, 2, 2)
    dfv = diff(f, v)
    x = None
    for a in range(3):
        for b in range(2):
            for c in range(2):
                self.assertEqual(dfv[a, b, c](x), g[a, b, c](x))


# --- Coefficient and argument input configurations


def test_single_scalar_coefficient_derivative(self):
    cell = triangle
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    u = Coefficient(space)
    v = TestFunction(space)
    a = 3 * u**2
    b = derivative(a, u, v)
    self.assertEqualAfterPreprocessing(b, 3 * (u * (2 * v)))


def test_single_vector_coefficient_derivative(self):
    cell = triangle
    V = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)
    u = Coefficient(space)
    v = TestFunction(space)
    a = 3 * dot(u, u)
    actual = derivative(a, u, v)
    expected = 3 * (2 * (u[i] * v[i]))
    assertEqualBySampling(actual, expected)


def test_multiple_coefficient_derivative(self):
    cell = triangle
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    W = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
    M = MixedElement([V, W])
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    w_space = FunctionSpace(domain, W)
    m_space = FunctionSpace(domain, M)
    uv = Coefficient(v_space)
    uw = Coefficient(w_space)
    v = TestFunction(m_space)
    vv, vw = split(v)

    a = sin(uv) * dot(uw, uw)

    actual = derivative(a, (uv, uw), split(v))
    expected = cos(uv) * vv * (uw[i] * uw[i]) + (uw[j] * vw[j]) * 2 * sin(uv)
    assertEqualBySampling(actual, expected)

    actual = derivative(a, (uv, uw), v)
    expected = cos(uv) * vv * (uw[i] * uw[i]) + (uw[j] * vw[j]) * 2 * sin(uv)
    assertEqualBySampling(actual, expected)


def test_indexed_coefficient_derivative(self):
    cell = triangle
    ident = Identity(2)
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    W = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    w_space = FunctionSpace(domain, W)
    u = Coefficient(w_space)
    v = TestFunction(v_space)

    w = dot(u, nabla_grad(u))
    # a = dot(w, w)
    a = (u[i] * u[k].dx(i)) * w[k]

    actual = derivative(a, u[0], v)

    dw = v * u[k].dx(0) + u[i] * ident[0, k] * v.dx(i)
    expected = 2 * w[k] * dw

    assertEqualBySampling(actual, expected)


def test_multiple_indexed_coefficient_derivative(self):
    cell = tetrahedron
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    V2 = MixedElement([V, V])
    W = FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
    v2_space = FunctionSpace(domain, V2)
    w_space = FunctionSpace(domain, W)
    u = Coefficient(w_space)
    w = Coefficient(w_space)
    v = TestFunction(v2_space)
    vu, vw = split(v)

    actual = derivative(cos(u[i] * w[i]), (u[2], w[1]), (vu, vw))
    expected = -sin(u[i] * w[i]) * (vu * w[2] + u[1] * vw)

    assertEqualBySampling(actual, expected)


def test_segregated_derivative_of_convection(self):
    cell = tetrahedron
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    W = FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", cell, 1, (3,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    w_space = FunctionSpace(domain, W)

    u = Coefficient(w_space)
    v = Coefficient(w_space)
    du = TrialFunction(v_space)
    dv = TestFunction(v_space)

    L = dot(dot(u, nabla_grad(u)), v)

    Lv = {}
    Lvu = {}
    for a in range(3):
        Lv[a] = derivative(L, v[a], dv)
        for b in range(3):
            Lvu[a, b] = derivative(Lv[a], u[b], du)

    for a in range(3):
        for b in range(3):
            form = Lvu[a, b] * dx
            fd = compute_form_data(form)
            pf = fd.preprocessed_form
            expand_indices(pf)

    k = Index()
    for a in range(3):
        for b in range(3):
            actual = Lvu[a, b]
            expected = du * u[a].dx(b) * dv + u[k] * du.dx(k) * dv
            assertEqualBySampling(actual, expected)


# --- User provided derivatives of coefficients


def test_coefficient_derivatives(self):
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)

    dv = TestFunction(space)

    f = Coefficient(space, count=0)
    g = Coefficient(space, count=1)
    df = Coefficient(space, count=2)
    dg = Coefficient(space, count=3)
    u = Coefficient(space, count=4)
    cd = {f: df, g: dg}

    integrand = inner(f, g)
    expected = (df * dv) * g + f * (dg * dv)

    F = integrand * dx
    J = derivative(F, u, dv, cd)
    fd = compute_form_data(J)
    actual = fd.preprocessed_form.integrals()[0].integrand()
    assert (actual * dx).signature() == (expected * dx).signature()
    self.assertEqual(replace(actual, fd.function_replace_map), expected)


def test_vector_coefficient_scalar_derivatives(self):
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    VV = FiniteElement("vector Lagrange", triangle, 1, (2,), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    vv_space = FunctionSpace(domain, VV)

    dv = TestFunction(v_space)

    df = Coefficient(vv_space, count=0)
    g = Coefficient(vv_space, count=1)
    f = Coefficient(vv_space, count=2)
    u = Coefficient(v_space, count=3)
    cd = {f: df}

    integrand = inner(f, g)

    i0, i1, i2, i3, i4 = [Index(count=c) for c in range(5)]
    expected = as_tensor(df[i1] * dv, (i1,))[i0] * g[i0]

    F = integrand * dx
    J = derivative(F, u, dv, cd)
    fd = compute_form_data(J)
    actual = fd.preprocessed_form.integrals()[0].integrand()
    assert (actual * dx).signature() == (expected * dx).signature()


def test_vector_coefficient_derivatives(self):
    V = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    VV = FiniteElement("Lagrange", triangle, 1, (2, 2), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    vv_space = FunctionSpace(domain, VV)

    dv = TestFunction(v_space)

    df = Coefficient(vv_space, count=0)
    g = Coefficient(v_space, count=1)
    f = Coefficient(v_space, count=2)
    u = Coefficient(v_space, count=3)
    cd = {f: df}

    integrand = inner(f, g)

    i0, i1, i2, i3, i4 = [Index(count=c) for c in range(5)]
    expected = as_tensor(df[i2, i1] * dv[i1], (i2,))[i0] * g[i0]

    F = integrand * dx
    J = derivative(F, u, dv, cd)
    fd = compute_form_data(J)
    actual = fd.preprocessed_form.integrals()[0].integrand()
    assert (actual * dx).signature() == (expected * dx).signature()
    # self.assertEqual(replace(actual, fd.function_replace_map), expected)


def test_vector_coefficient_derivatives_of_product(self):
    V = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    VV = FiniteElement("Lagrange", triangle, 1, (2, 2), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    vv_space = FunctionSpace(domain, VV)

    dv = TestFunction(v_space)

    df = Coefficient(vv_space, count=0)
    g = Coefficient(v_space, count=1)
    dg = Coefficient(vv_space, count=2)
    f = Coefficient(v_space, count=3)
    u = Coefficient(v_space, count=4)
    cd = {f: df, g: dg}

    integrand = f[i] * g[i]

    i0, i1, i2, i3, i4 = [Index(count=c) for c in range(5)]
    expected = (
        as_tensor(df[i2, i1] * dv[i1], (i2,))[i0] * g[i0]
        + f[i0] * as_tensor(dg[i4, i3] * dv[i3], (i4,))[i0]
    )

    F = integrand * dx
    J = derivative(F, u, dv, cd)
    fd = compute_form_data(J)
    actual = fd.preprocessed_form.integrals()[0].integrand()

    # Tricky case! These are equal in representation except
    # that the outermost sum/indexsum are swapped.
    # Sampling the expressions instead of comparing representations.
    x = (0, 0)
    funcs = {dv: (13, 14), f: (1, 2), g: (3, 4), df: ((5, 6), (7, 8)), dg: ((9, 10), (11, 12))}
    self.assertEqual(replace(actual, fd.function_replace_map)(x, funcs), expected(x, funcs))

    # TODO: Add tests covering more cases, in particular mixed stuff


# --- Some actual forms


def testHyperElasticity(self):
    cell = interval
    element = FiniteElement("Lagrange", cell, 2, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (1,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    w = Coefficient(space)
    v = TestFunction(space)
    u = TrialFunction(space)
    b = Constant(domain)
    K = Constant(domain)

    dw = w.dx(0)
    dv = v.dx(0)
    du = v.dx(0)

    E = dw + dw**2 / 2
    E = variable(E)
    Q = b * E**2
    psi = K * (exp(Q) - 1)

    f = psi * dx
    F = derivative(f, w, v)
    J = derivative(F, w, u)

    form_data_f = compute_form_data(f)
    form_data_F = compute_form_data(F)
    form_data_J = compute_form_data(J)

    f = form_data_f.preprocessed_form
    F = form_data_F.preprocessed_form
    J = form_data_J.preprocessed_form

    f_expression = strip_variables(f.integrals_by_type("cell")[0].integrand())
    F_expression = strip_variables(F.integrals_by_type("cell")[0].integrand())
    J_expression = strip_variables(J.integrals_by_type("cell")[0].integrand())

    # classes = set(c.__class__ for c in post_traversal(f_expression))

    Kv = 0.2
    bv = 0.3
    dw = 0.5
    dv = 0.7
    du = 0.11
    E = dw + dw**2 / 2.0
    Q = bv * E**2
    expQ = float(exp(Q))
    psi = Kv * (expQ - 1)
    fv = psi
    Fv = 2 * Kv * bv * E * (1 + dw) * expQ * dv
    Jv = 2 * Kv * bv * expQ * dv * du * (E + (1 + dw) ** 2 * (2 * bv * E**2 + 1))

    def Nv(x, derivatives):
        assert derivatives == (0,)
        return dv

    def Nu(x, derivatives):
        assert derivatives == (0,)
        return du

    def Nw(x, derivatives):
        assert derivatives == (0,)
        return dw

    mapping = {K: Kv, b: bv, w: Nw}
    fv2 = f_expression((0,), mapping)
    self.assertAlmostEqual(fv, fv2)

    (v,) = form_data_F.original_form.arguments()
    mapping = {K: Kv, b: bv, v: Nv, w: Nw}
    Fv2 = F_expression((0,), mapping)
    self.assertAlmostEqual(Fv, Fv2)

    v, u = form_data_J.original_form.arguments()
    mapping = {K: Kv, b: bv, v: Nv, u: Nu, w: Nw}
    Jv2 = J_expression((0,), mapping)
    self.assertAlmostEqual(Jv, Jv2)


def test_mass_derived_from_functional(self):
    cell = triangle
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)

    v = TestFunction(space)
    u = TrialFunction(space)
    w = Coefficient(space)

    f = (w**2 / 2) * dx
    L = w * v * dx
    # a = u*v*dx
    F = derivative(f, w, v)
    derivative(L, w, u)
    derivative(F, w, u)
    # TODO: assert something


# --- Interaction with replace


def test_derivative_replace_works_together(self):
    cell = triangle
    V = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, V)

    v = TestFunction(space)
    u = TrialFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)

    M = cos(f) * sin(g)
    F = derivative(M, f, v)
    J = derivative(F, f, u)
    JR = replace(J, {f: g})

    F2 = -sin(f) * v * sin(g)
    J2 = -cos(f) * u * v * sin(g)
    JR2 = -cos(g) * u * v * sin(g)

    assertEqualBySampling(F, F2)
    assertEqualBySampling(J, J2)
    assertEqualBySampling(JR, JR2)


def test_index_simplification_handles_repeated_indices(self):
    mesh = Mesh(FiniteElement("Lagrange", quadrilateral, 1, (2,), identity_pullback, H1))
    V = FunctionSpace(mesh, FiniteElement("DQ", quadrilateral, 0, (2, 2), identity_pullback, L2))
    K = JacobianInverse(mesh)
    G = outer(Identity(2), Identity(2))
    i, j, k, L, m, n = indices(6)
    A = as_tensor(K[m, i] * K[n, j] * G[i, j, k, L], (m, n, k, L))
    i, j = indices(2)
    # Can't use A[i, i, j, j] because UFL automagically index-sums
    # repeated indices in the __getitem__ call.
    Adiag = Indexed(A, MultiIndex((i, i, j, j)))
    A = as_tensor(Adiag, (i, j))
    v = TestFunction(V)
    f = inner(A, v) * dx
    fd = compute_form_data(f, do_apply_geometry_lowering=True)
    (integral,) = fd.preprocessed_form.integrals()
    assert integral.integrand().ufl_free_indices == ()


def test_index_simplification_reference_grad(self):
    mesh = Mesh(FiniteElement("Lagrange", quadrilateral, 1, (2,), identity_pullback, H1))
    (i,) = indices(1)
    A = as_tensor(Indexed(Jacobian(mesh), MultiIndex((i, i))), (i,))
    expr = apply_derivatives(apply_geometry_lowering(apply_algebra_lowering(A[0])))
    assert expr == ReferenceGrad(SpatialCoordinate(mesh))[0, 0]
    assert expr.ufl_free_indices == ()
    assert expr.ufl_shape == ()


# --- Scratch space


def test_foobar(self):
    element = FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    du = TrialFunction(space)
    U = Coefficient(space)

    def planarGrad(u):
        return as_matrix([[u[0].dx(0), 0, u[0].dx(1)], [0, 0, 0], [u[1].dx(0), 0, u[1].dx(1)]])

    def epsilon(u):
        return 0.5 * (planarGrad(u) + planarGrad(u).T)

    def NS_a(u, v):
        return inner(epsilon(u), epsilon(v))

    L = NS_a(U, v) * dx
    _ = derivative(L, U, du)
    # TODO: assert something



"""Test of expression comparison."""

from ufl import Coefficient, Cofunction, FunctionSpace, Mesh, triangle
from ufl.exprcontainers import ExprList
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_comparison_of_coefficients():
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    U = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)
    Ub = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    u_space = FunctionSpace(domain, U)
    ub_space = FunctionSpace(domain, Ub)

    v1 = Coefficient(v_space, count=1)
    v1b = Coefficient(v_space, count=1)
    v2 = Coefficient(v_space, count=2)
    u1 = Coefficient(u_space, count=1)
    u2 = Coefficient(u_space, count=2)
    u2b = Coefficient(ub_space, count=2)

    # Identical objects
    assert v1 == v1
    assert u2 == u2

    # Equal but distinct objects
    assert v1 == v1b
    assert u2 == u2b

    # Different objects
    assert not v1 == v2
    assert not u1 == u2
    assert not v1 == u1
    assert not v2 == u2


def test_comparison_of_cofunctions():
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    U = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)
    Ub = FiniteElement("Lagrange", triangle, 2, (), identity_pullback, H1)

    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    u_space = FunctionSpace(domain, U)
    ub_space = FunctionSpace(domain, Ub)

    v1 = Cofunction(v_space.dual(), count=1)
    v1b = Cofunction(v_space.dual(), count=1)
    v2 = Cofunction(v_space.dual(), count=2)
    u1 = Cofunction(u_space.dual(), count=1)
    u2 = Cofunction(u_space.dual(), count=2)
    u2b = Cofunction(ub_space.dual(), count=2)

    # Identical objects
    assert v1 == v1
    assert u2 == u2

    # Equal but distinct objects
    assert v1 == v1b
    assert u2 == u2b

    # Different objects
    assert not v1 == v2
    assert not u1 == u2
    assert not v1 == u1
    assert not v2 == u2

    # Objects in ExprList as happens when taking derivatives.
    assert ExprList(v1, v1) == ExprList(v1, v1b)
    assert not ExprList(v1, v2) == ExprList(v1, v1)


def test_comparison_of_products():
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    v = Coefficient(v_space)
    u = Coefficient(v_space)
    a = (v * 2) * u
    b = (2 * v) * u
    c = 2 * (v * u)
    assert a == b
    assert not a == c
    assert not b == c


def test_comparison_of_sums():
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    v = Coefficient(v_space)
    u = Coefficient(v_space)
    a = (v + 2) + u
    b = (2 + v) + u
    c = 2 + (v + u)
    assert a == b
    assert not a == c
    assert not b == c


def test_comparison_of_deeply_nested_expression():
    V = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    v_space = FunctionSpace(domain, V)
    v = Coefficient(v_space, count=1)
    u = Coefficient(v_space, count=1)
    w = Coefficient(v_space, count=2)

    def build_expr(a):
        for i in range(100):
            if i % 3 == 0:
                a = a + i
            elif i % 3 == 1:
                a = a * i
            elif i % 3 == 2:
                a = a**i
        return a

    a = build_expr(u)
    b = build_expr(v)
    c = build_expr(w)
    assert a == b
    assert not a == c
    assert not b == c



# Copyright (C) 2007 Anders Logg
#
# This file is part of UFL.
#
# UFL is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# UFL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with UFL. If not, see <http://www.gnu.org/licenses/>.
#
# Specification of a system F(v, u) = 0 and extraction of
# the bilinear and linear forms a and L for the left- and
# right-hand sides:
#
#   F(v, u) = a(v, u) - L(v) = 0
#
# The example below demonstrates the specification of the
# linear system for a cG(1)/Crank-Nicholson time step for
# the heat equation.
#
# The below formulation is equivalent to writing
#
#  a = v*u*dx + 0.5*k*dot(grad(v), grad(u))*dx
#  L = v*u0*dx - 0.5*k*dot(grad(v), grad(u0))*dx
#
# but instead of manually shuffling terms not including
# the unknown u to the right-hand side, all terms may
# be listed on one line and left- and right-hand sides
# extracted by lhs() and rhs().
from ufl import (
    Coefficient,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dot,
    dx,
    grad,
    lhs,
    rhs,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

k = 0.1

v = TestFunction(space)
u = TrialFunction(space)
u0 = Coefficient(space)

F = v * (u - u0) * dx + k * dot(grad(v), grad(0.5 * (u0 + u))) * dx

a = lhs(F)
L = rhs(F)



"""UFL type."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from __future__ import annotations

import typing
from abc import ABC, abstractmethod

import ufl.core as core
from ufl.core.compute_expr_hash import compute_expr_hash
from ufl.utils.formatting import camel2underscore


class UFLObject(ABC):
    """A UFL Object."""

    @abstractmethod
    def _ufl_hash_data_(self) -> typing.Hashable:
        """Return hashable data that uniquely defines this object."""

    @abstractmethod
    def __str__(self) -> str:
        """Return a human-readable string representation of the object."""

    @abstractmethod
    def __repr__(self) -> str:
        """Return a string representation of the object."""

    def __hash__(self) -> int:
        """Hash the object."""
        return hash(self._ufl_hash_data_())

    def __eq__(self, other):
        """Check if two objects are equal."""
        return type(self) is type(other) and self._ufl_hash_data_() == other._ufl_hash_data_()

    def __ne__(self, other):
        """Check inequality."""
        return not self.__eq__(other)


def get_base_attr(cls, name):
    """Return first non-``None`` attribute of given name among base classes."""
    for base in cls.mro():
        if hasattr(base, name):
            attr = getattr(base, name)
            if attr is not None:
                return attr
    return None


def set_trait(cls, basename, value, inherit=False):
    """Assign a trait to class with namespacing ``_ufl_basename_`` applied.

    If trait value is ``None``, optionally inherit it from the closest
    base class that has it.
    """
    name = "_ufl_" + basename + "_"
    if value is None and inherit:
        value = get_base_attr(cls, name)
    setattr(cls, name, value)


def determine_num_ops(cls, num_ops, unop, binop, rbinop):
    """Determine number of operands for this type."""
    # Try to determine num_ops from other traits or baseclass, or
    # require num_ops to be set for non-abstract classes if it cannot
    # be determined automatically
    if num_ops is not None:
        return num_ops
    elif cls._ufl_is_terminal_:
        return 0
    elif unop:
        return 1
    elif binop or rbinop:
        return 2
    else:
        # Determine from base class
        return get_base_attr(cls, "_ufl_num_ops_")


def check_is_terminal_consistency(cls):
    """Check for consistency in ``is_terminal`` trait among superclasses."""
    if cls._ufl_is_terminal_ is None:
        msg = (
            f"Class {cls.__name__} has not specified the is_terminal trait."
            " Did you forget to inherit from Terminal or Operator?"
        )
        raise TypeError(msg)

    base_is_terminal = get_base_attr(cls, "_ufl_is_terminal_")
    if base_is_terminal is not None and cls._ufl_is_terminal_ != base_is_terminal:
        msg = (
            f"Conflicting given and automatic 'is_terminal' trait for class {cls.__name__}."
            " Check if you meant to inherit from Terminal or Operator."
        )
        raise TypeError(msg)


def check_abstract_trait_consistency(cls):
    """Check that the first base classes up to ``Expr`` are other UFL types."""
    for base in cls.mro():
        if base is core.expr.Expr:
            break
        if not issubclass(base, core.expr.Expr) and base._ufl_is_abstract_:
            msg = (
                "Base class {0.__name__} of class {1.__name__} "
                "is not an abstract subclass of {2.__name__}."
            )
            raise TypeError(msg.format(base, cls, core.expr.Expr))


def check_has_slots(cls):
    """Check if type has __slots__ unless it is marked as exception with _ufl_noslots_."""
    if "_ufl_noslots_" in cls.__dict__:
        return

    if "__slots__" not in cls.__dict__:
        msg = (
            "Class {0.__name__} is missing the __slots__ "
            "attribute and is not marked with _ufl_noslots_."
        )
        raise TypeError(msg.format(cls))

    # Check base classes for __slots__ as well, skipping object which is the last one
    for base in cls.mro()[1:-1]:
        if "__slots__" not in base.__dict__:
            msg = "Class {0.__name__} is has a base class {1.__name__} with __slots__ missing."
            raise TypeError(msg.format(cls, base))


def check_type_traits_consistency(cls):
    """Execute a variety of consistency checks on the ufl type traits."""
    # Check for consistency in global type collection sizes
    Expr = core.expr.Expr
    assert Expr._ufl_num_typecodes_ == len(Expr._ufl_all_handler_names_)
    assert Expr._ufl_num_typecodes_ == len(Expr._ufl_all_classes_)
    assert Expr._ufl_num_typecodes_ == len(Expr._ufl_obj_init_counts_)
    assert Expr._ufl_num_typecodes_ == len(Expr._ufl_obj_del_counts_)

    # Check that non-abstract types always specify num_ops
    if not cls._ufl_is_abstract_:
        if cls._ufl_num_ops_ is None:
            msg = "Class {0.__name__} has not specified num_ops."
            raise TypeError(msg.format(cls))

    # Check for non-abstract types that num_ops has the right type
    if not cls._ufl_is_abstract_:
        if not (isinstance(cls._ufl_num_ops_, int) or cls._ufl_num_ops_ == "varying"):
            msg = 'Class {0.__name__} has invalid num_ops value {1} (integer or "varying").'
            raise TypeError(msg.format(cls, cls._ufl_num_ops_))

    # Check that num_ops is not set to nonzero for a terminal
    if cls._ufl_is_terminal_ and cls._ufl_num_ops_ != 0:
        msg = "Class {0.__name__} has num_ops > 0 but is terminal."
        raise TypeError(msg.format(cls))

    # Check that a non-scalar type doesn't have a scalar base class.
    if not cls._ufl_is_scalar_:
        if get_base_attr(cls, "_ufl_is_scalar_"):
            msg = "Non-scalar class {0.__name__} is has a scalar base class."
            raise TypeError(msg.format(cls))


def check_implements_required_methods(cls):
    """Check if type implements the required methods."""
    if not cls._ufl_is_abstract_:
        for attr in core.expr.Expr._ufl_required_methods_:
            if not hasattr(cls, attr):
                msg = "Class {0.__name__} has no {1} method."
                raise TypeError(msg.format(cls, attr))
            elif not callable(getattr(cls, attr)):
                msg = "Required method {1} of class {0.__name__} is not callable."
                raise TypeError(msg.format(cls, attr))


def check_implements_required_properties(cls):
    """Check if type implements the required properties."""
    if not cls._ufl_is_abstract_:
        for attr in core.expr.Expr._ufl_required_properties_:
            if not hasattr(cls, attr):
                msg = "Class {0.__name__} has no {1} property."
                raise TypeError(msg.format(cls, attr))
            elif callable(getattr(cls, attr)):
                msg = "Required property {1} of class {0.__name__} is a callable method."
                raise TypeError(msg.format(cls, attr))


def attach_implementations_of_indexing_interface(
    cls, inherit_shape_from_operand, inherit_indices_from_operand
):
    """Attach implementations of indexing interface."""
    # Scalar or index-free? Then we can simplify the implementation of
    # tensor properties by attaching them here.
    if cls._ufl_is_scalar_:
        cls.ufl_shape = ()

    if cls._ufl_is_scalar_ or cls._ufl_is_index_free_:
        cls.ufl_free_indices = ()
        cls.ufl_index_dimensions = ()

    # Automate direct inheriting of shape and indices from one of the
    # operands.  This simplifies refactoring because a lot of types do
    # this.
    if inherit_shape_from_operand is not None:

        def _inherited_ufl_shape(self):
            return self.ufl_operands[inherit_shape_from_operand].ufl_shape

        cls.ufl_shape = property(_inherited_ufl_shape)

    if inherit_indices_from_operand is not None:

        def _inherited_ufl_free_indices(self):
            return self.ufl_operands[inherit_indices_from_operand].ufl_free_indices

        def _inherited_ufl_index_dimensions(self):
            return self.ufl_operands[inherit_indices_from_operand].ufl_index_dimensions

        cls.ufl_free_indices = property(_inherited_ufl_free_indices)
        cls.ufl_index_dimensions = property(_inherited_ufl_index_dimensions)


def update_global_expr_attributes(cls):
    """Update global attributres.

    Update global ``Expr`` attributes, mainly by adding *cls* to global
    collections of ufl types.
    """
    if cls._ufl_is_terminal_modifier_:
        core.expr.Expr._ufl_terminal_modifiers_.append(cls)

    # Add to collection of language operators.  This collection is
    # used later to populate the official language namespace.
    # TODO: I don't think this functionality is fully completed, check
    # it out later.
    if not cls._ufl_is_abstract_ and hasattr(cls, "_ufl_function_"):
        cls._ufl_function_.__func__.__doc__ = cls.__doc__
        core.expr.Expr._ufl_language_operators_[cls._ufl_handler_name_] = cls._ufl_function_


def update_ufl_type_attributes(cls):
    """Update UFL type attributes."""
    # Determine integer typecode by incrementally counting all types
    cls._ufl_typecode_ = UFLType._ufl_num_typecodes_
    UFLType._ufl_num_typecodes_ += 1

    UFLType._ufl_all_classes_.append(cls)

    # Determine handler name by a mapping from "TypeName" to "type_name"
    cls._ufl_handler_name_ = camel2underscore(cls.__name__)
    UFLType._ufl_all_handler_names_.add(cls._ufl_handler_name_)

    # Append space for counting object creation and destriction of
    # this this type.
    UFLType._ufl_obj_init_counts_.append(0)
    UFLType._ufl_obj_del_counts_.append(0)


def ufl_type(
    is_abstract=False,
    is_terminal=None,
    is_scalar=False,
    is_index_free=False,
    is_shaping=False,
    is_literal=False,
    is_terminal_modifier=False,
    is_in_reference_frame=False,
    is_restriction=False,
    is_evaluation=False,
    is_differential=None,
    use_default_hash=True,
    num_ops=None,
    inherit_shape_from_operand=None,
    inherit_indices_from_operand=None,
    wraps_type=None,
    unop=None,
    binop=None,
    rbinop=None,
):
    """Decorator to apply to every subclass in the UFL ``Expr`` and ``BaseForm`` hierarchy.

    This decorator contains a number of checks that are intended to
    enforce uniform behaviour across UFL types.

    The rationale behind the checks and the meaning of the optional
    arguments should be sufficiently documented in the source code
    below.
    """

    def _ufl_type_decorator_(cls):
        """UFL type decorator."""
        # Update attributes for UFLType instances (BaseForm and Expr objects)
        update_ufl_type_attributes(cls)
        if not issubclass(cls, core.expr.Expr):
            # Don't need anything else for non Expr subclasses
            return cls

        # is_scalar implies is_index_freeg
        if is_scalar:
            _is_index_free = True
        else:
            _is_index_free = is_index_free

        # Store type traits
        cls._ufl_class_ = cls
        set_trait(cls, "is_abstract", is_abstract, inherit=False)

        set_trait(cls, "is_terminal", is_terminal, inherit=True)
        set_trait(cls, "is_literal", is_literal, inherit=True)
        set_trait(cls, "is_terminal_modifier", is_terminal_modifier, inherit=True)
        set_trait(cls, "is_shaping", is_shaping, inherit=True)
        set_trait(cls, "is_in_reference_frame", is_in_reference_frame, inherit=True)
        set_trait(cls, "is_restriction", is_restriction, inherit=True)
        set_trait(cls, "is_evaluation", is_evaluation, inherit=True)
        set_trait(cls, "is_differential", is_differential, inherit=True)

        set_trait(cls, "is_scalar", is_scalar, inherit=True)
        set_trait(cls, "is_index_free", _is_index_free, inherit=True)

        # Number of operands can often be determined automatically
        _num_ops = determine_num_ops(cls, num_ops, unop, binop, rbinop)
        set_trait(cls, "num_ops", _num_ops)

        # Attach builtin type wrappers to Expr
        """# These are currently handled in the as_ufl implementation in constantvalue.py
        if wraps_type is not None:
            if not isinstance(wraps_type, type):
                msg = "Expecting a type, not a {0.__name__} for the
                wraps_type argument in definition of {1.__name__}."
                raise TypeError(msg.format(type(wraps_type), cls))

            def _ufl_from_type_(value):
                return cls(value)
            from_type_name = "_ufl_from_{0}_".format(wraps_type.__name__)
            setattr(Expr, from_type_name, staticmethod(_ufl_from_type_))
        """

        # Attach special function to Expr.
        # Avoids the circular dependency problem of making
        # Expr.__foo__ return a Foo that is a subclass of Expr.
        """# These are currently attached in exproperators.py
        if unop:
            def _ufl_expr_unop_(self):
                return cls(self)
            setattr(Expr, unop, _ufl_expr_unop_)
        if binop:
            def _ufl_expr_binop_(self, other):
                try:
                    other = Expr._ufl_coerce_(other)
                except:
                    return NotImplemented
                return cls(self, other)
            setattr(Expr, binop, _ufl_expr_binop_)
        if rbinop:
            def _ufl_expr_rbinop_(self, other):
                try:
                    other = Expr._ufl_coerce_(other)
                except:
                    return NotImplemented
                return cls(other, self)
            setattr(Expr, rbinop, _ufl_expr_rbinop_)
        """

        # Make sure every non-abstract class has its own __hash__ and
        # __eq__.  Python 3 will set __hash__ to None if cls has
        # __eq__, but we've implemented it in a separate function and
        # want to inherit/use that for all types. Allow overriding by
        # setting use_default_hash=False.
        if use_default_hash:
            cls.__hash__ = compute_expr_hash

        # NB! This function conditionally adds some methods to the
        # class!  This approach significantly reduces the amount of
        # small functions to implement across all the types but of
        # course it's a bit more opaque.
        attach_implementations_of_indexing_interface(
            cls, inherit_shape_from_operand, inherit_indices_from_operand
        )

        # Update Expr
        update_global_expr_attributes(cls)

        # Apply a range of consistency checks to detect bugs in type
        # implementations that Python doesn't check for us, including
        # some checks that a static language compiler would do for us
        check_abstract_trait_consistency(cls)
        check_has_slots(cls)
        check_is_terminal_consistency(cls)
        check_implements_required_methods(cls)
        check_implements_required_properties(cls)
        check_type_traits_consistency(cls)

        return cls

    return _ufl_type_decorator_


class UFLType(type):
    """Base class for all UFL types.

    Equip UFL types with some ufl specific properties.
    """

    # A global counter of the number of typecodes assigned.
    _ufl_num_typecodes_ = 0

    # Set the handler name for UFLType
    _ufl_handler_name_ = "ufl_type"

    # A global array of all Expr and BaseForm subclasses, indexed by typecode
    _ufl_all_classes_ = []

    # A global set of all handler names added
    _ufl_all_handler_names_ = set()

    # A global array of the number of initialized objects for each
    # typecode
    _ufl_obj_init_counts_ = []

    # A global array of the number of deleted objects for each
    # typecode
    _ufl_obj_del_counts_ = []

    # Type trait: If the type is abstract.  An abstract class cannot
    # be instantiated and does not need all properties specified.
    _ufl_is_abstract_ = True

    # Type trait: If the type is terminal.
    _ufl_is_terminal_ = None



__authors__ = "Cecile Daversin Catty"
__date__ = "2019-03-26 -- 2019-03-26"

from ufl import (
    FunctionSpace,
    Measure,
    Mesh,
    MixedFunctionSpace,
    TestFunctions,
    TrialFunctions,
    interval,
    tetrahedron,
    triangle,
)
from ufl.algorithms.formsplitter import extract_blocks
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


def test_mixed_functionspace(self):
    # Domains
    domain_3d = Mesh(FiniteElement("Lagrange", tetrahedron, 1, (3,), identity_pullback, H1))
    domain_2d = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    domain_1d = Mesh(FiniteElement("Lagrange", interval, 1, (1,), identity_pullback, H1))
    # Finite elements
    f_1d = FiniteElement("Lagrange", interval, 1, (), identity_pullback, H1)
    f_2d = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    f_3d = FiniteElement("Lagrange", tetrahedron, 1, (), identity_pullback, H1)
    # Function spaces
    V_3d = FunctionSpace(domain_3d, f_3d)
    V_2d = FunctionSpace(domain_2d, f_2d)
    V_1d = FunctionSpace(domain_1d, f_1d)

    # MixedFunctionSpace = V_3d x V_2d x V_1d
    V = MixedFunctionSpace(V_3d, V_2d, V_1d)
    # Check sub spaces
    assert V.num_sub_spaces() == 3
    assert V.ufl_sub_space(0) == V_3d
    assert V.ufl_sub_space(1) == V_2d
    assert V.ufl_sub_space(2) == V_1d

    # Arguments from MixedFunctionSpace
    (u_3d, u_2d, u_1d) = TrialFunctions(V)
    (v_3d, v_2d, v_1d) = TestFunctions(V)

    # Measures
    dx3 = Measure("dx", domain=domain_3d)
    dx2 = Measure("dx", domain=domain_2d)
    dx1 = Measure("dx", domain=domain_1d)

    # Mixed variational form
    # LHS
    a_11 = u_1d * v_1d * dx1
    a_22 = u_2d * v_2d * dx2
    a_33 = u_3d * v_3d * dx3
    a_21 = u_2d * v_1d * dx1
    a_12 = u_1d * v_2d * dx1
    a_32 = u_3d * v_2d * dx2
    a_23 = u_2d * v_3d * dx2
    a_31 = u_3d * v_1d * dx1
    a_13 = u_1d * v_3d * dx1
    a = a_11 + a_22 + a_33 + a_21 + a_12 + a_32 + a_23 + a_31 + a_13
    # RHS
    f_1 = v_1d * dx1
    f_2 = v_2d * dx2
    f_3 = v_3d * dx3
    f = f_1 + f_2 + f_3

    # Check extract_block algorithm
    # LHS
    assert extract_blocks(a, 0, 0) == a_33
    assert extract_blocks(a, 0, 1) == a_23
    assert extract_blocks(a, 0, 2) == a_13
    assert extract_blocks(a, 1, 0) == a_32
    assert extract_blocks(a, 1, 1) == a_22
    assert extract_blocks(a, 1, 2) == a_12
    assert extract_blocks(a, 2, 0) == a_31
    assert extract_blocks(a, 2, 1) == a_21
    assert extract_blocks(a, 2, 2) == a_11
    # RHS
    assert extract_blocks(f, 0) == f_3
    assert extract_blocks(f, 1) == f_2
    assert extract_blocks(f, 2) == f_1

    # Test dual space method
    V_dual = V.dual()
    assert V_dual.num_sub_spaces() == 3
    assert V_dual.ufl_sub_space(0) == V_3d.dual()
    assert V_dual.ufl_sub_space(1) == V_2d.dual()
    assert V_dual.ufl_sub_space(2) == V_1d.dual()

    V_dual = V.dual(0, 2)
    assert V_dual.num_sub_spaces() == 3
    assert V_dual.ufl_sub_space(0) == V_3d.dual()
    assert V_dual.ufl_sub_space(1) == V_2d
    assert V_dual.ufl_sub_space(2) == V_1d.dual()



import os

import pytest

import ufl
from ufl import as_ufl, dx, inner
from ufl.algorithms import compute_form_data


class Tester:
    def assertTrue(self, a):
        assert a

    def assertFalse(self, a):
        assert not a

    def assertEqual(self, a, b):
        assert a == b

    def assertAlmostEqual(self, a, b):
        assert abs(a - b) < 1e-7

    def assertNotEqual(self, a, b):
        assert a != b

    def assertIsInstance(self, obj, cls):
        assert isinstance(obj, cls)

    def assertNotIsInstance(self, obj, cls):
        assert not isinstance(obj, cls)

    def assertRaises(self, e, f):
        assert pytest.raises(e, f)

    def assertEqualTotalShape(self, value, expected):
        self.assertEqual(value.ufl_shape, expected.ufl_shape)
        self.assertEqual(value.ufl_free_indices, expected.ufl_free_indices)
        self.assertEqual(value.ufl_index_dimensions, expected.ufl_index_dimensions)

    def assertSameIndices(self, expr, free_indices):
        self.assertEqual(expr.ufl_free_indices, tuple(sorted(i.count() for i in free_indices)))

    def assertEqualAfterPreprocessing(self, a, b):
        a2 = compute_form_data(a * dx).preprocessed_form
        b2 = compute_form_data(b * dx).preprocessed_form
        self.assertEqual(a2, b2)

    def assertEqualValues(self, A, B):
        B = as_ufl(B)
        self.assertEqual(A.ufl_shape, B.ufl_shape)
        self.assertEqual(inner(A - B, A - B)(None), 0)


@pytest.fixture(scope="session")
def self():
    return Tester()


def testspath():
    return os.path.abspath(os.path.dirname(__file__))


def filespath():
    return os.path.abspath(os.path.join(testspath(), "../demo"))


_all_cells = [ufl.interval, ufl.triangle, ufl.tetrahedron]


@pytest.fixture(params=_all_cells)
def cell(request):
    return request.param



"""Precedence handling."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import warnings

# FIXME: This code is crap...


def parstr(child, parent, pre="(", post=")", format=str):
    """Parstr."""
    # Execute when needed instead of on import, which leads to all
    # kinds of circular trouble.  Fixing this could be an optimization
    # of str(expr) though.
    if not hasattr(parent, "_precedence"):
        assign_precedences(build_precedence_list())

    # We want child to be evaluated fully first, and if the parent has
    # higher precedence we later wrap in ().
    s = format(child)

    # Operators where operands are always parenthesized because
    # precedence is not defined below
    if parent._precedence == 0:
        return pre + s + post

    # If parent operator binds stronger than child, must parenthesize
    # child
    # FIXME: Is this correct for all possible positions of () in a + b + c?
    # FIXME: Left-right rule
    if parent._precedence > child._precedence:  # parent = indexed, child = terminal
        return pre + s + post

    # Nothing needed
    return s


def build_precedence_list():
    """Build precedence list."""
    from ufl.classes import (
        Abs,
        BesselFunction,
        Division,
        Indexed,
        IndexSum,
        MathFunction,
        Operator,
        Power,
        Product,
        Sum,
        Terminal,
    )

    # TODO: Fill in other types...
    # Power <= Transposed

    precedence_list = []
    # Default operator behaviour: should always add parentheses
    precedence_list.append((Operator,))

    precedence_list.append((Sum,))

    # sum_i a + b = (sum_i a) + b != sum_i (a + b), sum_i binds
    # stronger than +, but weaker than product
    precedence_list.append((IndexSum,))

    precedence_list.append(
        (
            Product,
            Division,
        )
    )

    # NB! Depends on language!
    precedence_list.append((Power, MathFunction, BesselFunction, Abs))

    precedence_list.append((Indexed,))

    # Default terminal behaviour: should never add parentheses
    precedence_list.append((Terminal,))
    return precedence_list


def build_precedence_mapping(precedence_list):
    """Given a precedence list, build a dict with class->int mappings.

    Utility function used by some external code.
    """
    from ufl.classes import Expr, abstract_classes, all_ufl_classes

    pm = {}
    missing = set()
    # Assign integer values for each precedence level
    k = 0
    for p in precedence_list:
        for c in p:
            pm[c] = k
        k += 1
    # Check for missing classes, fill in subclasses
    for c in all_ufl_classes:
        if c not in abstract_classes and c not in pm:
            b = c.__bases__[0]
            while b is not Expr:
                if b in pm:
                    pm[c] = pm[b]
                    break
                b = b.__bases__[0]
            if c not in pm:
                missing.add(c)
    return pm, missing


def assign_precedences(precedence_list):
    """Given a precedence list, assign ints to class._precedence."""
    pm, missing = build_precedence_mapping(precedence_list)
    for c, p in sorted(pm.items(), key=lambda x: x[0].__name__):
        c._precedence = p
    if missing:
        warnings.warn(
            "Missing precedence levels for classes:\n"
            + "\n".join(f"  {c}" for c in sorted(missing))
        )



"""Check arities."""

from itertools import chain
from typing import Tuple

from ufl.classes import Argument, Zero
from ufl.corealg.map_dag import map_expr_dag
from ufl.corealg.multifunction import MultiFunction
from ufl.corealg.traversal import traverse_unique_terminals


class ArityMismatch(BaseException):
    """Arity mismatch exception."""

    pass


def _afmt(atuple: Tuple[Argument, bool]) -> str:
    """Return a string representation of an arity tuple."""
    arg, conj = atuple
    return f"conj({arg})" if conj else str(arg)


class ArityChecker(MultiFunction):
    """Arity checker."""

    def __init__(self, arguments):
        """Initialise."""
        MultiFunction.__init__(self)
        self.arguments = arguments
        self._et = ()

    def terminal(self, o):
        """Apply to terminal."""
        return self._et

    def argument(self, o):
        """Apply to argument."""
        return ((o, False),)

    def nonlinear_operator(self, o):
        """Apply to nonlinear_operator."""
        # Cutoff traversal by not having *ops in argument list of this
        # handler.  Traverse only the terminals under here the fastest
        # way we know of:
        for t in traverse_unique_terminals(o):
            if t._ufl_typecode_ == Argument._ufl_typecode_:
                raise ArityMismatch(
                    f"Applying nonlinear operator {o._ufl_class_.__name__} to "
                    f"expression depending on form argument {t}."
                )
        return self._et

    expr = nonlinear_operator

    def sum(self, o, a, b):
        """Apply to sum."""
        if a != b:
            raise ArityMismatch(
                f"Adding expressions with non-matching form arguments {_afmt(a)} vs {_afmt(b)}."
            )
        return a

    def division(self, o, a, b):
        """Apply to division."""
        if b:
            raise ArityMismatch(f"Cannot divide by form argument {b}.")
        return a

    def product(self, o, a, b):
        """Apply to product."""
        if a and b:
            # Check that we don't have test*test, trial*trial, even
            # for different parts in a block system
            anumbers = set(x[0].number() for x in a)
            for x in b:
                if x[0].number() in anumbers:
                    raise ArityMismatch(
                        "Multiplying expressions with overlapping form argument number "
                        f"{x[0].number()}, argument is {_afmt(x)}."
                    )
            # Combine argument lists
            c = tuple(sorted(set(a + b), key=lambda x: (x[0].number(), x[0].part())))
            # Check that we don't have any arguments shared between a
            # and b
            if len(c) != len(a) + len(b) or len(c) != len({x[0] for x in c}):
                raise ArityMismatch(
                    "Multiplying expressions with overlapping form arguments "
                    f"{_afmt(a)} vs {_afmt(b)}."
                )
            # It's fine for argument parts to overlap
            return c
        elif a:
            return a
        else:
            return b

    # inner, outer and dot all behave as product but for conjugates
    def inner(self, o, a, b):
        """Apply to inner."""
        return self.product(o, a, self.conj(None, b))

    dot = inner

    def outer(self, o, a, b):
        """Apply to outer."""
        return self.product(o, self.conj(None, a), b)

    def linear_operator(self, o, a):
        """Apply to linear_operator."""
        return a

    # Positive and negative restrictions behave as linear operators
    positive_restricted = linear_operator
    negative_restricted = linear_operator

    # Cell and facet average are linear operators
    cell_avg = linear_operator
    facet_avg = linear_operator

    # Grad is a linear operator
    grad = linear_operator
    reference_grad = linear_operator
    reference_value = linear_operator

    # Conj, is a sesquilinear operator
    def conj(self, o, a):
        """Apply to conj."""
        return tuple((a_[0], not a_[1]) for a_ in a)

    # Does it make sense to have a Variable(Argument)? I see no
    # problem.
    def variable(self, o, f, a):
        """Apply to variable."""
        return f

    # Conditional is linear on each side of the condition
    def conditional(self, o, c, a, b):
        """Apply to conditional."""
        if c:
            raise ArityMismatch(f"Condition cannot depend on form arguments ({_afmt(a)}).")
        if a and isinstance(o.ufl_operands[2], Zero):
            # Allow conditional(c, arg, 0)
            return a
        elif b and isinstance(o.ufl_operands[1], Zero):
            # Allow conditional(c, 0, arg)
            return b
        elif a == b:
            # Allow conditional(c, test, test)
            return a
        else:
            # Do not allow e.g. conditional(c, test, trial),
            # conditional(c, test, nonzeroconstant)
            raise ArityMismatch(
                "Conditional subexpressions with non-matching form arguments "
                f"{_afmt(a)} vs {_afmt(b)}."
            )

    def linear_indexed_type(self, o, a, i):
        """Apply to linear_indexed_type."""
        return a

    # All of these indexed thingies behave as a linear_indexed_type
    indexed = linear_indexed_type
    index_sum = linear_indexed_type
    component_tensor = linear_indexed_type

    def list_tensor(self, o, *ops):
        """Apply to list_tensor."""
        args = set(chain(*ops))
        if args:
            # Check that each list tensor component has the same
            # argument numbers (ignoring parts)
            numbers = set(tuple(sorted(set(arg[0].number() for arg in op))) for op in ops)
            if () in numbers:  # Allow e.g. <v[0], 0, v[1]> but not <v[0], u[0]>
                numbers.remove(())
            if len(numbers) > 1:
                raise ArityMismatch(
                    "Listtensor components must depend on the same argument numbers, "
                    f"found {numbers}."
                )

            # Allow different parts with the same number
            return tuple(sorted(args, key=lambda x: (x[0].number(), x[0].part())))
        else:
            # No argument dependencies
            return self._et


def check_integrand_arity(expr, arguments, complex_mode=False):
    """Check the arity of an integrand."""
    arguments = tuple(sorted(set(arguments), key=lambda x: (x.number(), x.part())))
    rules = ArityChecker(arguments)
    arg_tuples = map_expr_dag(rules, expr, compress=False)
    args = tuple(a[0] for a in arg_tuples)
    if args != arguments:
        raise ArityMismatch(f"Integrand arguments {args} differ from form arguments {arguments}.")
    if complex_mode:
        # Check that the test function is conjugated and that any
        # trial function is not conjugated. Further arguments are
        # treated as trial funtions (i.e. no conjugation) but this
        # might not be correct.
        for arg, conj in arg_tuples:
            if arg.number() == 0 and not conj:
                raise ArityMismatch("Failure to conjugate test function in complex Form")
            elif arg.number() > 0 and conj:
                raise ArityMismatch(f"Argument {arg} is spuriously conjugated in complex Form")


def check_form_arity(form, arguments, complex_mode=False):
    """Check the arity of a form."""
    for itg in form.integrals():
        check_integrand_arity(itg.integrand(), arguments, complex_mode)



"""Basic algorithms for applying functions to sub-expressions."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

# NOTE: Placing this under algorithms/ because I want corealg/ to stay clean
# as part of a careful refactoring process, and this file depends on ufl.form
# which drags in a lot of stuff.

from ufl.action import Action
from ufl.adjoint import Adjoint
from ufl.constantvalue import Zero
from ufl.core.expr import Expr
from ufl.corealg.map_dag import map_expr_dag
from ufl.form import BaseForm, Form, FormSum, ZeroBaseForm
from ufl.integral import Integral


def map_integrands(function, form, only_integral_type=None):
    """Map integrands.

    Apply transform(expression) to each integrand expression in form, or
    to form if it is an Expr.
    """
    if isinstance(form, Form):
        mapped_integrals = [
            map_integrands(function, itg, only_integral_type) for itg in form.integrals()
        ]
        nonzero_integrals = [
            itg for itg in mapped_integrals if not isinstance(itg.integrand(), Zero)
        ]
        return Form(nonzero_integrals)
    elif isinstance(form, Integral):
        itg = form
        if (only_integral_type is None) or (itg.integral_type() in only_integral_type):
            return itg.reconstruct(function(itg.integrand()))
        else:
            return itg
    elif isinstance(form, FormSum):
        mapped_components = [
            map_integrands(function, component, only_integral_type)
            for component in form.components()
        ]
        nonzero_components = [
            (component, w)
            for component, w in zip(mapped_components, form.weights())
            # Catch ufl.Zero and ZeroBaseForm
            if component != 0
        ]

        # Simplify case with one nonzero component and the corresponding weight is 1
        if len(nonzero_components) == 1 and nonzero_components[0][1] == 1:
            return nonzero_components[0][0]

        if all(not isinstance(component, BaseForm) for component, _ in nonzero_components):
            # Simplification of `BaseForm` objects may turn `FormSum` into a sum of `Expr` objects
            # that are not `BaseForm`, i.e. into a `Sum` object.
            # Example: `Action(Adjoint(c*), u)` with `c*` a `Coargument` and u a `Coefficient`.
            return sum([component for component, _ in nonzero_components])
        return FormSum(*nonzero_components)
    elif isinstance(form, Adjoint):
        # Zeros are caught inside `Adjoint.__new__`
        return Adjoint(map_integrands(function, form._form, only_integral_type))
    elif isinstance(form, Action):
        left = map_integrands(function, form._left, only_integral_type)
        right = map_integrands(function, form._right, only_integral_type)
        # Zeros are caught inside `Action.__new__`
        return Action(left, right)
    elif isinstance(form, ZeroBaseForm):
        arguments = tuple(
            map_integrands(function, arg, only_integral_type) for arg in form._arguments
        )
        return ZeroBaseForm(arguments)
    elif isinstance(form, (Expr, BaseForm)):
        integrand = form
        return function(integrand)
    else:
        raise ValueError("Expecting Form, Integral or Expr.")


def map_integrand_dags(function, form, only_integral_type=None, compress=True):
    """Map integrand dags."""
    return map_integrands(
        lambda expr: map_expr_dag(function, expr, compress), form, only_integral_type
    )



"""This module provides utility functions for computing permutations and generating index lists."""
# Copyright (C) 2008-2016 Anders Logg and Kent-Andre Mardal
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Martin Alnæs 2009-2016


def compute_indices(shape):
    """Compute all index combinations for given shape."""
    if len(shape) == 0:
        return ((),)
    sub_indices = compute_indices(shape[1:])
    indices = []
    for i in range(shape[0]):
        for sub_index in sub_indices:
            indices.append((i,) + sub_index)
    return tuple(indices)


def build_component_numbering(shape, symmetry):
    """Build a numbering of components within the given value shape.

    This takes into consideration a symmetry mapping which leaves the
    mapping noncontiguous. Returns a dict { component -> numbering }
    and an ordered list of components [ numbering -> component ].  The
    dict contains all components while the list only contains the ones
    not mapped by the symmetry mapping.
    """
    vi2si, si2vi = {}, []
    indices = compute_indices(shape)
    # Number components not in symmetry mapping
    for c in indices:
        if c not in symmetry:
            vi2si[c] = len(si2vi)
            si2vi.append(c)
    # Copy numbering to mapped components
    for c in indices:
        if c in symmetry:
            vi2si[c] = vi2si[symmetry[c]]
    # Validate
    for k, c in enumerate(si2vi):
        assert vi2si[c] == k
    return vi2si, si2vi



"""Sobolev spaces.

This module defines a symbolic heirarchy of Sobolev spaces to enable
symbolic reasoning about the spaces in which finite elements lie.
"""
# Copyright (C) 2014 Imperial College London and others
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by David Ham 2014
#
# Modified by Martin Alnaes 2014
# Modified by Lizao Li 2015
# Modified by Thomas Gibson 2017

from functools import total_ordering
from math import inf, isinf

__all_classes__ = ["SobolevSpace", "DirectionalSobolevSpace"]


@total_ordering
class SobolevSpace(object):
    """Symbolic representation of a Sobolev space.

    This implements a subset of the methods of a Python set so that
    finite elements and other Sobolev spaces can be tested for
    inclusion.
    """

    def __init__(self, name, parents=None):
        """Instantiate a SobolevSpace object.

        Args:
            name: The name of this space,
            parents: A set of Sobolev spaces of which this
            space is a subspace.
        """
        self.name = name
        p = frozenset(parents or [])
        # Ensure that the inclusion operations are transitive.
        self.parents = p.union(*[p_.parents for p_ in p])
        self._order = {
            "L2": 0,
            "H1": 1,
            "H2": 2,
            "HInf": inf,
            # Order for the elements below is taken from
            # its parent Sobolev space
            "HDiv": 0,
            "HCurl": 0,
            "HEin": 0,
            "HDivDiv": 0,
            "DirectionalH": 0,
        }[self.name]

    def __str__(self):
        """Format as a string."""
        return self.name

    def __repr__(self):
        """Representation."""
        return f"SobolevSpace({self.name!r}, {list(self.parents)!r})"

    def __eq__(self, other):
        """Check equality."""
        return isinstance(other, SobolevSpace) and self.name == other.name

    def __ne__(self, other):
        """Not equal."""
        return not self == other

    def __hash__(self):
        """Hash."""
        return hash(("SobolevSpace", self.name))

    def __getitem__(self, spatial_index):
        """Returns the Sobolev space associated with a particular spatial coordinate."""
        return self

    def __contains__(self, other):
        """Implement `fe in s` where `fe` is a FiniteElement and `s` is a SobolevSpace."""
        if isinstance(other, SobolevSpace):
            raise TypeError(
                "Unable to test for inclusion of a SobolevSpace in another SobolevSpace. "
                "Did you mean to use <= instead?"
            )
        return other.sobolev_space == self or self in other.sobolev_space.parents

    def __lt__(self, other):
        """In common with intrinsic Python sets, < indicates "is a proper subset of"."""
        return other in self.parents


@total_ordering
class DirectionalSobolevSpace(SobolevSpace):
    """Directional Sobolev space.

    Symbolic representation of a Sobolev space with varying smoothness
    in different spatial directions.
    """

    def __init__(self, orders):
        """Instantiate a DirectionalSobolevSpace object.

        Args:
            orders: an iterable of orders of weak derivatives, where
                the position denotes in what spatial variable the
                smoothness requirement is enforced.
        """
        assert all(
            isinstance(x, int) or isinf(x) for x in orders
        ), "Order must be an integer or infinity."
        name = "DirectionalH"
        parents = [L2]
        super(DirectionalSobolevSpace, self).__init__(name, parents)
        self._orders = tuple(orders)
        self._spatial_indices = range(len(self._orders))

    def __getitem__(self, spatial_index):
        """Returns the Sobolev space associated with a particular spatial coordinate."""
        if spatial_index not in range(len(self._orders)):
            raise IndexError("Spatial index out of range.")
        spaces = {0: L2, 1: H1, 2: H2, inf: HInf}
        return spaces[self._orders[spatial_index]]

    def __contains__(self, other):
        """Check if one space is contained in another.

        Implement `fe in s` where `fe` is a FiniteElement and `s` is a
        DirectionalSobolevSpace.
        """
        if isinstance(other, SobolevSpace):
            raise TypeError(
                "Unable to test for inclusion of a SobolevSpace in another SobolevSpace. "
                "Did you mean to use <= instead?"
            )
        return other.sobolev_space == self or all(
            self[i] in other.sobolev_space.parents for i in self._spatial_indices
        )

    def __eq__(self, other):
        """Check equality."""
        if isinstance(other, DirectionalSobolevSpace):
            return self._orders == other._orders
        return all(self[i] == other for i in self._spatial_indices)

    def __lt__(self, other):
        """In common with intrinsic Python sets, < indicates "is a proper subset of."""
        if isinstance(other, DirectionalSobolevSpace):
            if self._spatial_indices != other._spatial_indices:
                return False
            return any(self._orders[i] > other._orders[i] for i in self._spatial_indices)

        if other in [HDiv, HCurl]:
            return all(self._orders[i] >= 1 for i in self._spatial_indices)
        elif other.name in ["HDivDiv", "HEin"]:
            # Don't know how these spaces compare
            return NotImplementedError(f"Don't know how to compare with {other.name}")
        else:
            return any(self._orders[i] > other._order for i in self._spatial_indices)

    def __str__(self):
        """Format as a string."""
        return f"{self.name}({', '.join(map(str, self._orders))})"


L2 = SobolevSpace("L2")
HDiv = SobolevSpace("HDiv", [L2])
HCurl = SobolevSpace("HCurl", [L2])
H1 = SobolevSpace("H1", [HDiv, HCurl, L2])
H2 = SobolevSpace("H2", [H1, HDiv, HCurl, L2])
HInf = SobolevSpace("HInf", [H2, H1, HDiv, HCurl, L2])
HEin = SobolevSpace("HEin", [L2])
HDivDiv = SobolevSpace("HDivDiv", [L2])



"""Algorithms for estimating polynomial degrees of expressions."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs and Anders Logg
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Anders Logg, 2009-2010
# Modified by Jan Blechta, 2012

import warnings

from ufl.checks import is_cellwise_constant
from ufl.constantvalue import IntValue
from ufl.corealg.map_dag import map_expr_dags
from ufl.corealg.multifunction import MultiFunction
from ufl.domain import extract_unique_domain
from ufl.form import Form
from ufl.integral import Integral


class SumDegreeEstimator(MultiFunction):
    """Sum degree estimator.

    This algorithm is exact for a few operators and heuristic for many.
    """

    def __init__(self, default_degree, element_replace_map):
        """Initialise."""
        MultiFunction.__init__(self)
        self.default_degree = default_degree
        self.element_replace_map = element_replace_map

    def constant_value(self, v):
        """Apply to constant_value.

        Constant values are constant.
        """
        return 0

    def constant(self, v):
        """Apply to constant."""
        return 0

    def geometric_quantity(self, v):
        """Apply to geometric_quantity.

        Some geometric quantities are cellwise constant. Others are
        nonpolynomial and thus hard to estimate.
        """
        if is_cellwise_constant(v):
            return 0
        else:
            # As a heuristic, just returning domain degree to bump up degree somewhat
            return extract_unique_domain(v).ufl_coordinate_element().embedded_superdegree

    def spatial_coordinate(self, v):
        """Apply to spatial_coordinate.

        A coordinate provides additional degrees depending on coordinate field of domain.
        """
        return extract_unique_domain(v).ufl_coordinate_element().embedded_superdegree

    def cell_coordinate(self, v):
        """Apply to cell_coordinate.

        A coordinate provides one additional degree.
        """
        return 1

    def argument(self, v):
        """Apply to argument.

        A form argument provides a degree depending on the element,
        or the default degree if the element has no degree.
        """
        return (
            v.ufl_element().embedded_superdegree
        )  # FIXME: Use component to improve accuracy for mixed elements

    def coefficient(self, v):
        """Apply to coefficient.

        A form argument provides a degree depending on the element,
        or the default degree if the element has no degree.
        """
        e = v.ufl_element()
        e = self.element_replace_map.get(e, e)
        d = e.embedded_superdegree  # FIXME: Use component to improve accuracy for mixed elements
        if d is None:
            d = self.default_degree
        return d

    def _reduce_degree(self, v, f):
        """Reduce the estimated degree by one.

        This is used when derivatives are taken. It does not reduce the degree when
        TensorProduct elements or quadrilateral elements are involved.
        """
        if isinstance(f, int) and extract_unique_domain(v).ufl_cell().cellname() not in [
            "quadrilateral",
            "hexahedron",
        ]:
            return max(f - 1, 0)
        else:
            return f

    def _add_degrees(self, v, *ops):
        """Apply to _add_degrees."""
        if any(isinstance(o, tuple) for o in ops):
            # we can add a slight hack here to handle things
            # like adding 0 to (3, 3) [by expanding
            # 0 to (0, 0) when making tempops]
            tempops = [foo if isinstance(foo, tuple) else (foo, foo) for foo in ops]
            return tuple(map(sum, zip(*tempops)))
        else:
            return sum(ops)

    def _max_degrees(self, v, *ops):
        """Apply to _max_degrees."""
        if any(isinstance(o, tuple) for o in ops):
            tempops = [foo if isinstance(foo, tuple) else (foo, foo) for foo in ops]
            return tuple(map(max, zip(*tempops)))
        else:
            return max(ops + (0,))

    def _not_handled(self, v, *args):
        """Apply to _not_handled."""
        raise ValueError(f"Missing degree handler for type {v._ufl_class_.__name__}")

    def expr(self, v, *ops):
        """Apply to expr.

        For most operators we take the max degree of its operands.
        """
        warnings.warn(f"Missing degree estimation handler for type {v._ufl_class_.__name__}")
        return self._add_degrees(v, *ops)

    # Utility types with no degree concept
    def multi_index(self, v):
        """Apply to multi_index."""
        return None

    def label(self, v):
        """Apply to label."""
        return None

    # Fall-through, indexing and similar types
    def reference_value(self, rv, f):
        """Apply to reference_value."""
        return f

    def variable(self, v, e, a):
        """Apply to variable."""
        return e

    def transposed(self, v, A):
        """Apply to transposed."""
        return A

    def index_sum(self, v, A, ii):
        """Apply to index_sum."""
        return A

    def indexed(self, v, A, ii):
        """Apply to indexed."""
        return A

    def component_tensor(self, v, A, ii):
        """Apply to component_tensor."""
        return A

    list_tensor = _max_degrees

    def positive_restricted(self, v, a):
        """Apply to positive_restricted."""
        return a

    def negative_restricted(self, v, a):
        """Apply to negative_restricted."""
        return a

    def conj(self, v, a):
        """Apply to conj."""
        return a

    def real(self, v, a):
        """Apply to real."""
        return a

    def imag(self, v, a):
        """Apply to imag."""
        return a

    # A sum takes the max degree of its operands:
    sum = _max_degrees

    # TODO: Need a new algorithm which considers direction of
    # derivatives of form arguments A spatial derivative reduces the
    # degree with one
    grad = _reduce_degree
    reference_grad = _reduce_degree
    # Handling these types although they should not occur... please
    # apply preprocessing before using this algorithm:
    nabla_grad = _reduce_degree
    div = _reduce_degree
    reference_div = _reduce_degree
    nabla_div = _reduce_degree
    curl = _reduce_degree
    reference_curl = _reduce_degree

    def cell_avg(self, v, a):
        """Apply to cell_avg.

        Cell average of a function is always cellwise constant.
        """
        return 0

    def facet_avg(self, v, a):
        """Apply to facet_avg.

        Facet average of a function is always cellwise constant.
        """
        return 0

    # A product accumulates the degrees of its operands:
    product = _add_degrees
    # Handling these types although they should not occur... please
    # apply preprocessing before using this algorithm:
    inner = _add_degrees
    dot = _add_degrees
    outer = _add_degrees
    cross = _add_degrees

    # Explicitly not handling these types, please apply preprocessing
    # before using this algorithm:
    derivative = _not_handled  # base type
    compound_derivative = _not_handled  # base type
    compound_tensor_operator = _not_handled  # base class
    variable_derivative = _not_handled
    trace = _not_handled
    determinant = _not_handled
    cofactor = _not_handled
    inverse = _not_handled
    deviatoric = _not_handled
    skew = _not_handled
    sym = _not_handled

    def abs(self, v, a):
        """Apply to abs.

        This is a heuristic, correct if there is no.
        """
        if a == 0:
            return a
        else:
            return a

    def division(self, v, *ops):
        """Apply to division.

        Using the sum here is a heuristic. Consider e.g. (x+1)/(x-1).
        """
        return self._add_degrees(v, *ops)

    def power(self, v, a, b):
        """Apply to power.

        If b is a positive integer: degree(a**b) == degree(a)*b
        otherwise use the heuristic: degree(a**b) == degree(a) + 2.
        """
        f, g = v.ufl_operands

        if isinstance(g, IntValue):
            gi = g.value()
            if gi >= 0:
                if isinstance(a, int):
                    return a * gi
                else:
                    return tuple(foo * gi for foo in a)

        # Something to a non-(positive integer) power, e.g. float,
        # negative integer, Coefficient, etc.
        return self._add_degrees(v, a, 2)

    def atan2(self, v, a, b):
        """Apply to atan2.

        Using the heuristic:
        degree(atan2(const,const)) == 0
        degree(atan2(a,b)) == max(degree(a),degree(b))+2
        which can be wildly inaccurate but at least gives a somewhat
        high integration degree.
        """
        if a or b:
            return self._add_degrees(v, self._max_degrees(v, a, b), 2)
        else:
            return self._max_degrees(v, a, b)

    def math_function(self, v, a):
        """Apply to math_function.

        Using the heuristic:
        degree(sin(const)) == 0
        degree(sin(a)) == degree(a)+2
        which can be wildly inaccurate but at least gives a somewhat
        high integration degree.
        """
        if a:
            return self._add_degrees(v, a, 2)
        else:
            return a

    def bessel_function(self, v, nu, x):
        """Apply to bessel_function.

        Using the heuristic
        degree(bessel_*(const)) == 0
        degree(bessel_*(x)) == degree(x)+2
        which can be wildly inaccurate but at least gives a somewhat
        high integration degree.
        """
        if x:
            return self._add_degrees(v, x, 2)
        else:
            return x

    def condition(self, v, *args):
        """Apply to condition."""
        return None

    def conditional(self, v, c, t, f):
        """Apply to conditional.

        Degree of condition does not influence degree of values which conditional takes. So
        heuristicaly taking max of true degree and false degree. This will be exact in cells
        where condition takes single value. For improving accuracy of quadrature near
        condition transition surface quadrature order must be adjusted manually.
        """
        return self._max_degrees(v, t, f)

    def min_value(self, v, a, r):
        """Apply to min_value.

        Same as conditional.
        """
        return self._max_degrees(v, a, r)

    max_value = min_value

    def coordinate_derivative(self, v, integrand_degree, b, direction_degree, d):
        """Apply to coordinate_derivative.

        We use the heuristic that a shape derivative in direction V
        introduces terms V and grad(V) into the integrand. Hence we add the
        degree of the deformation to the estimate.
        """
        return self._add_degrees(v, integrand_degree, direction_degree)

    def expr_list(self, v, *o):
        """Apply to expr_list."""
        return self._max_degrees(v, *o)

    def expr_mapping(self, v, *o):
        """Apply to expr_mapping."""
        return self._max_degrees(v, *o)


def estimate_total_polynomial_degree(e, default_degree=1, element_replace_map={}):
    """Estimate total polynomial degree of integrand.

    NB: Although some compound types are supported here,
    some derivatives and compounds must be preprocessed
    prior to degree estimation. In generic code, this algorithm
    should only be applied after preprocessing.

    For coefficients defined on an element with unspecified degree
    (None), the degree is set to the given default degree.
    """
    de = SumDegreeEstimator(default_degree, element_replace_map)
    if isinstance(e, Form):
        if not e.integrals():
            raise ValueError("Form has no integrals.")
        degrees = map_expr_dags(de, [it.integrand() for it in e.integrals()])
    elif isinstance(e, Integral):
        degrees = map_expr_dags(de, [e.integrand()])
    else:
        degrees = map_expr_dags(de, [e])
    degree = max(degrees) if degrees else default_degree
    return degree



#
# Author: Martin Sandve Alnes
# Date: 2008-10-03
#
from ufl import Coefficient, FunctionSpace, Mesh, dx, triangle
from ufl.finiteelement import FiniteElement
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1

element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
space = FunctionSpace(domain, element)

f = Coefficient(space)

a = f**2 * dx



import pytest

from ufl import (
    Coefficient,
    Cofunction,
    Form,
    FormSum,
    FunctionSpace,
    Mesh,
    SpatialCoordinate,
    TestFunction,
    TrialFunction,
    dot,
    ds,
    dx,
    grad,
    inner,
    nabla_grad,
    triangle,
)
from ufl.finiteelement import FiniteElement
from ufl.form import BaseForm
from ufl.pullback import identity_pullback
from ufl.sobolevspace import H1


@pytest.fixture
def element():
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    return element


@pytest.fixture
def domain():
    cell = triangle
    return Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))


@pytest.fixture
def mass(domain):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    return u * v * dx


@pytest.fixture
def stiffness(domain):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    return inner(grad(u), grad(v)) * dx


@pytest.fixture
def convection(domain):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1)
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    u = TrialFunction(space)
    w = Coefficient(space)
    return dot(dot(w, nabla_grad(u)), v) * dx


@pytest.fixture
def load(domain):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    space = FunctionSpace(domain, element)
    f = Coefficient(space)
    v = TestFunction(space)
    return f * v * dx


@pytest.fixture
def boundary_load(domain):
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    space = FunctionSpace(domain, element)
    f = Coefficient(space)
    v = TestFunction(space)
    return f * v * ds


def test_form_arguments(mass, stiffness, convection, load):
    v, u = mass.arguments()
    (f,) = load.coefficients()

    assert v.number() == 0
    assert u.number() == 1
    assert stiffness.arguments() == (v, u)
    assert load.arguments() == (v,)

    assert (v * dx).arguments() == (v,)
    assert (v * dx + v * ds).arguments() == (v,)
    assert (v * dx + f * v * ds).arguments() == (v,)
    assert (u * v * dx(1) + v * u * dx(2)).arguments() == (v, u)
    assert ((f * v) * u * dx + (u * 3) * (v / 2) * dx(2)).arguments() == (v, u)


def test_form_coefficients(element, domain):
    space = FunctionSpace(domain, element)
    v = TestFunction(space)
    f = Coefficient(space)
    g = Coefficient(space)

    assert (g * dx).coefficients() == (g,)
    assert (g * dx + g * ds).coefficients() == (g,)
    assert (g * dx + f * ds).coefficients() == (f, g)
    assert (g * dx(1) + f * dx(2)).coefficients() == (f, g)
    assert (g * v * dx + f * v * dx(2)).coefficients() == (f, g)


def test_form_domains():
    cell = triangle
    element = FiniteElement("Lagrange", cell, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", cell, 1, (2,), identity_pullback, H1))
    V = FunctionSpace(domain, element)

    v = TestFunction(V)
    f = Coefficient(V)
    x = SpatialCoordinate(domain)[0]

    assert (x * dx).ufl_domains() == (domain,)
    assert (v * dx).ufl_domains() == (domain,)
    assert (f * dx).ufl_domains() == (domain,)
    assert (x * v * f * dx).ufl_domains() == (domain,)
    assert (1 * dx(domain)).ufl_domains() == (domain,)


def test_form_empty(mass):
    assert not mass.empty()
    assert Form([]).empty()


def test_form_integrals(mass, boundary_load):
    assert isinstance(mass.integrals(), tuple)
    assert len(mass.integrals()) == 1
    assert mass.integrals()[0].integral_type() == "cell"
    assert mass.integrals_by_type("cell") == mass.integrals()
    assert mass.integrals_by_type("exterior_facet") == ()
    assert isinstance(boundary_load.integrals_by_type("cell"), tuple)
    assert len(boundary_load.integrals_by_type("cell")) == 0
    assert len(boundary_load.integrals_by_type("exterior_facet")) == 1


def test_form_call():
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    V = FunctionSpace(domain, element)
    v = TestFunction(V)
    u = TrialFunction(V)
    f = Coefficient(V)
    g = Coefficient(V)
    a = g * inner(grad(v), grad(u)) * dx
    M = a(f, f, coefficients={g: 1})
    assert M == grad(f) ** 2 * dx

    import sys

    if sys.version_info.major >= 3 and sys.version_info.minor >= 5:
        a = u * v * dx
        M = eval("(a @ f) @ g")
        assert M == g * f * dx


def test_formsum(mass):
    element = FiniteElement("Lagrange", triangle, 1, (), identity_pullback, H1)
    domain = Mesh(FiniteElement("Lagrange", triangle, 1, (2,), identity_pullback, H1))
    V = FunctionSpace(domain, element)
    v = Cofunction(V.dual())

    assert v + mass
    assert mass + v
    assert isinstance((mass + v), FormSum)

    assert len((mass + v + v).components()) == 3
    # Variational forms are summed appropriately
    assert len((mass + v + mass).components()) == 2

    assert v - mass
    assert mass - v
    assert isinstance((mass + v), FormSum)

    assert -v
    assert isinstance(-v, BaseForm)
    assert (-v).weights()[0] == -1

    assert 2 * v
    assert isinstance(2 * v, BaseForm)
    assert (2 * v).weights()[0] == 2



"""Predicates for recognising duals."""
# Copyright (C) 2021 India Marsden
#
# This file is part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#


def is_primal(object):
    """Determine if the object belongs to a primal space.

    This is not simply the negation of is_dual,
    because a mixed function space containing both primal
    and dual components is neither primal nor dual.
    """
    return hasattr(object, "_primal") and object._primal


def is_dual(object):
    """Determine if the object belongs to a dual space.

    This is not simply the negation of is_primal,
    because a mixed function space containing both primal
    and dual components is neither primal nor dual.
    """
    return hasattr(object, "_dual") and object._dual



import pytest

import ufl


def test_interval():
    cell = ufl.interval
    assert cell.num_vertices() == 2
    assert cell.num_edges() == 1
    assert cell.num_faces() == 0


def test_triangle():
    cell = ufl.triangle
    assert cell.num_vertices() == 3
    assert cell.num_edges() == 3
    assert cell.num_faces() == 1


def test_quadrilateral():
    cell = ufl.quadrilateral
    assert cell.num_vertices() == 4
    assert cell.num_edges() == 4
    assert cell.num_faces() == 1


def test_tetrahedron():
    cell = ufl.tetrahedron
    assert cell.num_vertices() == 4
    assert cell.num_edges() == 6
    assert cell.num_faces() == 4


def test_hexahedron():
    cell = ufl.hexahedron
    assert cell.num_vertices() == 8
    assert cell.num_edges() == 12
    assert cell.num_faces() == 6


def test_pentatope():
    cell = ufl.pentatope
    assert cell.num_vertices() == 5
    assert cell.num_edges() == 10
    assert cell.num_faces() == 10


def test_tesseract():
    cell = ufl.tesseract
    assert cell.num_vertices() == 16
    assert cell.num_edges() == 32
    assert cell.num_faces() == 24


@pytest.mark.parametrize("cell", [ufl.interval])
def test_cells_1d(cell):
    assert cell.num_facets() == cell.num_vertices()
    assert cell.num_ridges() == 0
    assert cell.num_peaks() == 0


@pytest.mark.parametrize("cell", [ufl.triangle, ufl.quadrilateral])
def test_cells_2d(cell):
    assert cell.num_facets() == cell.num_edges()
    assert cell.num_ridges() == cell.num_vertices()
    assert cell.num_peaks() == 0


@pytest.mark.parametrize("cell", [ufl.tetrahedron, ufl.hexahedron, ufl.prism, ufl.pyramid])
def test_cells_3d(cell):
    assert cell.num_facets() == cell.num_faces()
    assert cell.num_ridges() == cell.num_edges()
    assert cell.num_peaks() == cell.num_vertices()


@pytest.mark.parametrize("cell", [ufl.tesseract, ufl.pentatope])
def test_cells_4d(cell):
    assert cell.num_facets() == cell.num_sub_entities(3)
    assert cell.num_ridges() == cell.num_faces()
    assert cell.num_peaks() == cell.num_edges()


def test_tensorproductcell():
    orig = ufl.TensorProductCell(ufl.interval, ufl.interval)
    cell = orig.reconstruct()
    assert cell.sub_cells() == orig.sub_cells()
    assert cell.topological_dimension() == orig.topological_dimension()
